from nbautoeval import ExerciseGenerator, GeneratorArgs, IsliceRenderer, PPrintRenderer

# usual function that returns a generator
def squares(n):
    for i in range(n):
        yield i**2
        
squares_inputs = [
    GeneratorArgs(2),
    GeneratorArgs(5, islice=(1, 4)),
    GeneratorArgs(3, islice=(5,)),
    GeneratorArgs(10, islice=(2, 8, 3)),
    GeneratorArgs(4, islice=(4,)),
    GeneratorArgs(5, islice=(4,)),
    GeneratorArgs(10, islice=(0, 10)),
    GeneratorArgs(10, islice=(0, 5)),
    GeneratorArgs(10, islice=(0, 15)),
    GeneratorArgs(10, islice=(5, 15)),
    GeneratorArgs(10, islice=(4, 10, 4)),
    GeneratorArgs(10, islice=(4, 20, 4)),
    GeneratorArgs(10, islice=(5, 15, 5)),
    GeneratorArgs(10, islice=(5, 15, 3)),
    GeneratorArgs(10, islice=(15, 20)),
]


exo_squares = ExerciseGenerator(
    squares, squares_inputs, nb_examples=4,
    result_renderer=PPrintRenderer(width=30),
)


### same with max_iterations
exo_squares_maxed = ExerciseGenerator(
    squares, squares_inputs, 
    nb_examples=0, max_iterations=5,
    result_renderer=PPrintRenderer(width=30),
)


###
from itertools import count

simple_inputs = [
    GeneratorArgs(),
    GeneratorArgs(islice=(10, 12)),
]

exo_count_maxed = ExerciseGenerator(
    count, simple_inputs,
    nb_examples=0, max_iterations=5,
    result_renderer=PPrintRenderer(width=30),
)



# -*- coding: utf-8 -*-

# pylint: disable=c0111, r1705, w0703


import re

from .exercise_function import ExerciseFunction

DEFAULT_MATCH_MODE = 'match'

class ExerciseRegexp(ExerciseFunction):
    """
    With these exercises the students are asked to write a regexp
    which is transformed into a function that essentially
    takes an input string and returns a boolean
    that says if the *whole* string matches or not
    """
    
    def regexp_to_solution(self, regexp, match_mode):
        def solution(string):
            if match_mode in ('match', 'search'):
                if match_mode == 'match':
                    match = re.match(regexp, string)
                else:
                    match = re.search(regexp, string)
                if not match:
                    return False
                else:
                    # xxx this is wrong ! matching section
                    # won't have the metachars in incoming pattern
                    return match.group(0) == string
            # findall returns strings, while finditer returns match instances
            elif match_mode == 'findall':
                return re.findall(regexp, string)
            elif match_mode == 'finditer':
                return [match.span()
                        for match in re.finditer(regexp, string)]
            return None
        return solution

    def __init__(self, name, regexp, inputs,
                 *args, match_mode=DEFAULT_MATCH_MODE, **keywords):
        """
        a regexp exercise is made with
        . a user-friendly name
        . a regexp pattern for the solution
        . a list of inputs on which to run the regexp
        . match_mode is either 'match', 'search' or 'findall'
        . additional settings from ExerciseFunction
        """
        solution = self.regexp_to_solution(regexp, match_mode)
        super().__init__(solution, inputs, *args, **keywords)
        self.regexp = regexp
        self.name = name
        self.match_mode = match_mode
        # it never makes sense to show the function name
        # that would always be 'solution' anyways
        self.call_renderer.show_function = False

    def correction(self, student_regexp):               # pylint: disable=w0221
        student_solution = self.regexp_to_solution(student_regexp, self.match_mode)
        return ExerciseFunction.correction(self, student_solution)
    
    @property
    def column_headers(self):
        return (self._column_headers if self._column_headers is not None 
            else ('chaîne', 'match ?', 'obtenu'))
    

##############################
class ExerciseRegexpGroups(ExerciseRegexp):
    """
    With these exercises the students are asked to write a regexp
    with a set of specified named groups
    a list of these groupnames needs to be passed to construct the object

    the regexp is then transformed into a function that again
    takes an input string and either a list of tuples
    (groupname, found_substring)
    or None if no match occurs
    """

    @staticmethod
    def extract_group(match, group):
        try:
            return group, match.group(group)
        except Exception:
            return group, "Undefined"

    def regexp_to_solution(self, regexp, match_mode):
        groups = self.groups
        if match_mode != 'match':
            # only tested with 'match' so far
            print(f"WARNING: ExerciseRegexpGroups : "
                  f"match_mode {match_mode} not yet implemented")
        def solution(string):
            if match_mode in ('match', 'search'):
                if match_mode == 'match':
                    match = re.match(regexp, string)
                else:
                    match = re.search(regexp, string)
                return match and [ExerciseRegexpGroups.extract_group(match, group)
                                  for group in groups]
            # findall returns strings, while finditer returns match instances
            elif match_mode == 'findall':
                return re.findall(regexp, string)
            elif match_mode == 'finditer':
                matches = re.finditer(regexp, string)
                return [[ExerciseRegexpGroups.extract_group(match, group)
                         for group in groups] for match in matches]
            return None
        return solution

    def __init__(self, name, regexp, groups, inputs,
                 *args, match_mode=DEFAULT_MATCH_MODE, **keywords):
        self.groups = groups
        super().__init__(name, regexp, inputs, *args, match_mode=match_mode, **keywords)

    @property
    def column_headers(self):
        return (self._column_headers if self._column_headers is not None 
            else ('chaîne', 'groupes', 'obtenu'))



# -*- coding: utf-8 -*-

# pylint: disable=c0111, w0703, r1705

############################################################
# the low level interface - used to be used directly in the first exercises

from ipywidgets import GridBox, Layout

from .content import TextContent, CssContent, ResultContent
from .callrenderer import Call, CallRenderer
from .renderer import Renderer
from .helpers import default_font_size, default_header_font_size
from .storage import log_correction, log2_correction


DEBUG = False
#DEBUG = True


default_column_headers_visible_function_name = (
    "appel",
    "attendu",
    "obtenu",
)
default_column_headers_no_function_name = (
    "arguments",
    "attendu",
    "obtenu",
)
# same order
column_span_classes = (
    "", "", "span-3-to-4",
)

###
CSS = """
.nbae-fun {
    width: max-content;
    border-top: .1em solid black;
    border-bottom: .1em solid black;
    padding: 0.1em 0em;
    row-gap: 0.5em;
    column-gap: .4em;
}
.nbae-fun .header {
    padding: .6em;
    background-color: #f1fcfc;
    border-bottom: .1em solid #666;
    font-weight: bold;
    border-top-left-radius: .4em;
    border-top-right-radius: .4em;
}
.nbae-fun .span-3-to-4 {
    grid-column: 3 / span 2;
}
div.nbae-fun pre {
    padding: 0px;
    line-height: 1.15;
    max-width: 100%;
    overflow: hidden;
}
div.nbae-fun .cell span {
    line-height: 1.15;
}
.nbae-fun .cell.ok, .nbae-fun .cell.ok code, .nbae-fun .result.ok {
    background-color: #d4f8e8;
}
.nbae-fun .cell.ok.even, .nbae-fun .cell.ok code.even, .nbae-fun .result.ok.even {
    background-color: #c4f8d8;
}
.nbae-fun .cell.ko, .nbae-fun .cell.ko code, .nbae-fun .result.ko {
    background-color: #ffc6c9;
}
.nbae-fun .cell.ko.even, .nbae-fun .cell.ko code.even, .nbae-fun .result.ko.even {
    background-color: #ffb6b9;
}
.nbae-fun .cell {
    padding: .2em .4em;
    border-radius: .4em;
    background-color: #f8f8f8;
    max-width: 100%;
}
.nbae-fun .cell.even {
    background-color: #f0f0f0;
}
.nbae-fun .ko.result {
    font-weight: bold;
    font-size: '130%';
}
.nbae-fun .result {
    padding: 4px;
}
.nbae-fun div.widget-html-content {
    display: flex;
}
"""

####################
class ExerciseFunction:                                           # pylint: disable=r0902
    """The class for an exercise where students are asked to write a
    function The teacher version of that function is provided as
    'solution' and is used against datasets to generate an online
    correction or example.

    A dataset is typically an instance of Args, it describes
    how the function is to be called (the arguments, both named and unnamed).

    **NOTE** that the Args model predates keyword-only and a fortiori
    positional-only parameters, so these might be a bit awkward to play with.

    The most useful method in this class is 'correction'; for each
    input in the dataset, we call both the teacher function and the
    student function, and compare the results using '==' to produce a
    table of green or red cells.

    The class provides a few other utility methods, notably 'example'
    that can be used in the students notebook to show the expected
    result for some or all of the inputs.

    One important aspects of this is copying. Realizing that both
    teacher and student functions can do side effects in the inputs,
    it means that these need to be copied before any call is made. By
    default the copy is a deep copy, but for some corner cases it can
    be required to use shallow copy instead; in this case just pass
    copy_mode='shallow' to the constructor here.

    In terms of rendering, an ExerciseFunction object requires 2 renderer objects

    * call_renderer is used to compute the contents of the leftmost column in the output
      of both correction() and example();
      this attribute should be a `CallRenderer` instance, that has a `render()` method
      that works on `Call` instances, and that returns a `Content` object.
    * result_renderer works similarly, it is used to render the results of the function
      calls in the internal columns of the output of correction(), and the rightmost
      column of the output of example();

    Typical uses of these 2 rendering attributes would be

    * ExerciseFunction(call_renderer=CallRenderer(show_function=False))
      allows to remove the function name, showing only arguments, to save space
    * ExerciseFunction(call_renderer=PPrintRenderer(width=30))
    * ExerciseFunction(result_renderer=PPrintRenderer(width=30)) allows to set
      a fixed limit limit for either calls or results.
    """
    def __init__(self, solution, datasets,              # pylint: disable=r0913
                 *,
                 copy_mode='deep',
                 nb_examples=1,
                 # how to render
                 call_renderer=None,
                 result_renderer=None,
                 #
                 column_headers=None,
                 font_size=default_font_size,
                 header_font_size=default_header_font_size,
                 # used for reports/logs, defaults computed from solution
                 name=None):
        # the 'official' solution
        self.solution = solution
        # the inputs - actually Args instances
        self.datasets = datasets
        # how to copy args
        self.copy_mode = copy_mode
        # how many examples
        self.nb_examples = nb_examples
        # renderers
        self.call_renderer = call_renderer or CallRenderer()
        self.result_renderer = result_renderer or Renderer()
        # header names: at this point, just remember any data passed
        # it's too erly to compute the actual value, as show_function
        # could be turned off later on - see e.g. ExerciseRegexp
        self._column_headers = column_headers
        #
        self.header_font_size = header_font_size
        self.font_size = font_size
        ###
        # in some weird cases this won't exist
        self.name = name or getattr(solution, '__name__', "no_name")


    @property
    def column_headers(self):
        return (self._column_headers if self._column_headers is not None
            else default_column_headers_visible_function_name if self.call_renderer.show_function
            else default_column_headers_no_function_name)


    def correction(self, student_function):             # pylint: disable=r0914
        """
        colums should be a 3-tuple for the 3 columns widths
        copy_mode can be either None, 'shallow', or 'deep' (default)
        or 'tee' for generators
        """
        datasets = self.datasets
        copy_mode = self.copy_mode

        #
        headers_props = {'font-size': self.header_font_size}
        body_props = {'font-size': self.font_size}
        contents = [TextContent(x, css_properties=headers_props)
                    .add_classes(['header', span_class])
                    for (x, span_class) in zip(self.column_headers, column_span_classes)]


        overall = True

        for index, dataset in enumerate(datasets):
            # will use original dataset for rendering to avoid any side-effects
            # during running

            # always clone all inputs
            if copy_mode != 'tee':
                student_dataset = dataset.clone(copy_mode)
                ref_dataset = dataset.clone(copy_mode)
            else:
                student_dataset, ref_dataset = dataset.copy_for_tee('tee')

            # run both codes
            ref_exc, stu_exc = False, False
            try:
                expected = ref_dataset.call(self.solution, debug=DEBUG)
            except Exception as exc:
                expected = exc
                ref_exc = True

            try:
                student_result = student_dataset.call(student_function, debug=DEBUG)
            except Exception as exc:
                student_result = exc
                stu_exc = True

            # compare results
            is_ok = self.validate(expected, student_result)
            if not is_ok:
                overall = False
            # render that run
            result_content = ResultContent(is_ok)
            classes = ['cell']
            if index % 2 == 0:
                classes.append("even")

            call = Call(self.solution, dataset)
            contents.append(self.call_renderer.render(call)
                            .add_classes(classes)
                            .add_css_properties(body_props))
            contents.append(self.result_renderer.render(expected)
                            .add_classes(classes).add_class('ok')
                            .add_css_properties(body_props)
                            .set_is_code(not ref_exc))
            contents.append(self.result_renderer.render(student_result)
                            .add_classes(classes)
                            .add_class(result_content.the_class())
                            .add_css_properties(body_props)
                            .set_is_code(not stu_exc))
            contents.append(result_content
                            .add_classes(classes)
                            .add_css_properties(body_props))

        log_correction(self.name, overall)
        # xxx would make sense to expose how many examples were right or wrong
        log2_correction(self.name, success=overall)

        contents.append(CssContent(CSS))

        gridbox_layout  = Layout(grid_template_columns = 'max-content 1fr 1fr max-content',
                                 max_width="100%")
        grid = GridBox([content.widget() for content in contents],
                       layout=gridbox_layout).add_class("nbae-fun")
        return grid


    # public interface
    def example(self, how_many=None):

        if how_many is None:
            how_many = self.nb_examples
        if how_many == 0:
            how_many = len(self.datasets)


        headers_props = {'font-size': self.header_font_size}
        body_props = {'font-size': self.font_size}
        contents = [TextContent(x, css_properties=headers_props).add_class('header')
                    for x in self.column_headers[:2]]
        #

        for index, dataset in enumerate(self.datasets[:how_many]):
            # clone enve in example() mode to avoid altering datasets
            if self.copy_mode != 'tee':
                sample_dataset = dataset.clone(self.copy_mode)
            else:
                sample_dataset, dataset = dataset.copy_for_tee(self.copy_mode)

            # run
            try:
                expected = sample_dataset.call(self.solution)
            except Exception as exc:                     # pylint:disable=w0703
                expected = exc

            # render that row
            classes = ['cell', 'example']
            if index % 2 == 0:
                classes.append("even")


            call = Call(self.solution, dataset)
            contents.append(self.call_renderer.render(call)
                            .add_classes(classes)
                            .add_css_properties(body_props))
            contents.append(self.result_renderer.render(expected)
                            .add_classes(classes)
                            .add_css_properties(body_props))

        contents.append(CssContent(CSS))

        gridbox_layout  = Layout(grid_template_columns = 'max-content 1fr')
        grid = GridBox([content.widget() for content in contents],
                       layout=gridbox_layout).add_class("nbae-fun")
        return grid


    def validate(self, expected, result):               # pylint: disable=r0201
        """
        how to compare the results as obtained from
        * the solution function
        * and the student function

        the default here is to use ==
        """

        if DEBUG:
            print(f"ExerciseFunction.validate is comparing {expected} with {result}")

        return expected == result

# see this question on SO
# https://stackoverflow.com/questions/40659212/futurewarning-elementwise-comparison-failed-returning-scalar-but-in-the-futur

try:
    import numpy as np
    import warnings

    class ExerciseFunctionNumpy(ExerciseFunction):
        """
        This is suitable for functions that are expected to return a numpy (nd)array
        """

        def __init__(self, solution, datasets,
                     *args,
                     **kwds):
            ExerciseFunction.__init__(
                self, solution, datasets,
                # xxx check this again
                *args, **kwds)

        # redefine validation function on numpy arrays
        def validate(self, expected, result):
            try:
                return np.all(
                    np.isclose(
                        expected, result))
            except Exception:
                # print("OOPS", type(e), e)
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter(action='ignore', category=FutureWarning)
                        # we need to return a genuine bool here
                        result = expected == result
                        if isinstance(result, np.ndarray):
                            return np.all(result)
                        else:
                            return result
                except Exception as exc:
                    print("OOPS2", type(exc), exc)
                    return False

except Exception:
    #print("ExerciseFunctionNumpy not defined ; numpy not installed ? ")
    pass



# -*- coding: utf-8 -*-
from nbautoeval import ExerciseRegexp, ExerciseRegexpGroups
from nbautoeval.args import Args

########## step 1
# one correct answer

# @BEG@ week=6 sequence=6 name=pythonid more=regexp
at_least_two = "TA(TA)+"
# @END@

########## step 2 : the datasets

# used for generating all orders
import itertools

germs = [ 'AGCT', 'CAGT', 'TATG', 'CGTA' ]

# building the inputs
inputs =  []
# one run will be made on this string as-is
inputs += ['AGCTTATATATACAGT']

# here we construct 24 strings by combining the 4 seeds above in all possible orders
inputs += [
    "".join(x) for x in itertools.permutations(germs)
    ]
# so this amounts to 25 different runs

########## step 3
exo_at_least_two = ExerciseRegexp(
    # for building a regexp we need to provide a name explicitly
    # (as opposed to when we deal with functions, that have an internal name)
    'at_least_two',
    # the pattern that describes the correct regexp
    at_least_two,
    # the datasets are, as always, a list of Args objects
    [Args(x) for x in inputs],
    # match_mode can be either 'match' or 'search', depending on
    match_mode = 'finditer',
    nb_examples = 8)

# XXX to be clarified again
at_least_two_ko = "TATA"




from types import FunctionType, BuiltinFunctionType, BuiltinMethodType

default_font_size='small'
default_header_font_size='medium'


# helpers 
def commas(iterable):
    if isinstance(iterable, dict):
        return ", ".join(f"{k}={custom_repr(v)}" for k, v in iterable.items())
    elif isinstance(iterable, str):
        return str
    else:
        return ", ".join([custom_repr(x) for x in iterable])


def custom_repr(x):
    if isinstance(x, (type, FunctionType, BuiltinFunctionType, BuiltinMethodType)):
        return x.__name__
    elif isinstance(x, set):
        return "{" + commas(x) + "}"
    else:
        return repr(x)


def truncate(message: str, width):
    # width <= 0 means do not truncate
    if width <= 0:
        return message
    truncated = message if len(message) <= width \
        else message[:width-3]+'...'
    return truncated



# -*- coding: utf-8 -*-

# pylint: disable=c0111, r1705, w0703

import itertools
import functools
from collections.abc import Iterator

from .exercise_function import ExerciseFunction
from .callrenderer import IsliceRenderer


class ExerciseGenerator(ExerciseFunction):
    """
    With these exercises the students are asked to write a generator
    function; the Args mechanism is the same as for regular functions;
    an optional 'max_iterations' attribute allows to deal with infinite 
    iterators
    validation is done by 
    (*) counting the number of items in the enumeration 
        maxed by max_iterations if relevant
    (*) comparing each of the results yielded by next()
    """

    @staticmethod
    def generator_to_solution(generator_function, max_iterations=None):
        @functools.wraps(generator_function)
        def solution(*args, **kwds):
            # call the function written by the student
            generator = generator_function(*args, **kwds)
            if not isinstance(generator, (Iterator, range)):
                raise TypeError(f"not an iterator! received a {type(generator).__name__} instance: {generator} ")
            if max_iterations is None:
                return list(generator)
            result = list(itertools.islice(generator, None, max_iterations+1))
            if len(result) == max_iterations+1:
                result[-1] = '...'
            return result
        return solution


    def __init__(self, generator_function, datasets, max_iterations=None,
                 *args, **keywords):
        """
        a generator exercise is made with
        . a generator function for the solution
        . a list of Args instances to produce iterators that are then tested
        . max_iterations is a global limit on the number of items 
          that are attempted to be retrieved
        . additional settings from ExerciseFunction
        """
        # change default
        solution = ExerciseGenerator.generator_to_solution(generator_function,
                                                           max_iterations)
        if 'call_renderer' not in keywords:
            keywords['call_renderer'] = IsliceRenderer()
        super().__init__(solution, datasets, *args, **keywords)
        self.generator_function = generator_function
        self.max_iterations = max_iterations
        # that was part of the aborted 0.6.1 attempt
        # to copy incoming generators - see also issue #4
        #if 'copy_mode' not in keywords:
        #    self.copy_mode = 'tee'


    def correction(self, student_generator):                    
        student_solution = ExerciseGenerator.generator_to_solution(
            student_generator, self.max_iterations)
        return ExerciseFunction.correction(self, student_solution)





from nbautoeval import ExerciseClass, Args, ClassScenario, ClassExpression, ClassStatement


# provide a workable class

####################
class Fifo:

    def __init__(self):
        self.items = []
        
    # see the check_init flag below 
    def __repr__(self):
        contents = ", ".join(str(item) for item in self.items)
        return f"[Fifo {contents}]"

    def incoming(self, incoming):
        self.items.append(incoming)

    def outgoing(self):
        if len(self.items) == 0:
            return None
        return self.items.pop(0)
    
    def __len__(self):
        return len(self.items)


#####
# define 2 flavours of the same exercise

# I: for newbies
# if like me you like to have students code their own stack the very first day
# at a point where __repr__() is not yet mastered, then the way ExerciseFunction
# by default checks for objects status cannot work well, since it does rely on repr()
# 
# you can still come up with the simplest assignment ever, 
# using check_init=True that by-passes checks after the first (initialization) step
# you must refrain from using Statements though in this case, because Statements
# are also checked using repr() 


exo_fifo_newbies = ExerciseClass(
    Fifo,
    [
        ClassScenario(
            Args(),
            ClassExpression("INSTANCE.incoming(1)"),
            ClassExpression("INSTANCE.outgoing()"),
        ),
        ClassScenario(
            Args(),
            ClassExpression("INSTANCE.incoming(1)"),
            ClassExpression("INSTANCE.incoming(2)"),
            ClassExpression("INSTANCE.outgoing()"),
            ClassExpression("INSTANCE.outgoing()"),
        ),
        ],
    check_init=False, nb_examples=0)
    
    

# a more realistic one
# this one will require the student to write a repr() 
# that exactly matches the official implementation

scenario1 = ClassScenario(
    # arguments to the constructor
    Args(),
    # a list of expressions, with 
    # INSTANCE and CLASS replaced as appropriate
    ClassExpression("INSTANCE.incoming(1)"),
    ClassExpression("INSTANCE.incoming(2)"),
    ClassExpression("INSTANCE"),
    # same than "INSTANCE" but with a slightly different display
    ClassExpression("repr(INSTANCE)"),
    ClassStatement("INSTANCE"),
    ClassExpression("INSTANCE.outgoing()"),
    ClassExpression("INSTANCE.incoming(3)"),
    ClassExpression("INSTANCE.incoming(4)"),
    ClassExpression("INSTANCE.outgoing()"),
    ClassExpression("len(INSTANCE)"),
)

scenario2 = ClassScenario(
    Args(),
    "INSTANCE.outgoing()",
    "INSTANCE.incoming(1)",
    "INSTANCE.incoming(2)",
    "INSTANCE.incoming(3)",
    "INSTANCE.incoming(4)",
    "INSTANCE.outgoing()",
)    

exo_fifo = ExerciseClass (Fifo, [scenario1, scenario2], check_init=False)

if __name__ == '__main__':
    exo_fifo.correction(Fifo)




import pprint

from .content import TextContent

r"""
various ways to render a Python object

a (result) Renderer is expected to provide a render() method on all python objects

in turn render() is expected to return a "Content"
"""


class Renderer:
    """
    default renderer:
    * on classes that own a `_render_content_` method, use that;
      it should return a TextContent
    * otherwise use repr()
    """

    def __init__(self):
        pass

    def render(self, python_object):
        if hasattr(python_object, "_render_content_"):
            return python_object._render_content_()
        else:
            return (TextContent(repr(python_object))
                    .add_css_properties({'align-self': 'center'})
                    .set_is_code(not isinstance(python_object, Exception)))


    def __repr__(self):
        return f"<Renderer {type(self)}>"


class PPrintRenderer(Renderer):
    # the default for compact with pprint is False, but we favour True
    def __init__(self, *, width=80, indent=2, compact=True):
        self.width = width
        self.indent = indent
        self.compact = compact

    def render(self, python_object):
        return (TextContent(
                    pprint.pformat(python_object, compact=self.compact,
                                   indent=self.indent, width=self.width))
                    .add_css_properties({'align-self': 'center'})
                    .set_is_code(True))


class MultilineRenderer(Renderer):
    def render(self, python_object):
        if not isinstance(python_object, str):
            python_object = repr(python_object)
        return TextContent("\\n\n".join(python_object.split("\n")))



import pprint

from .args import ArgsTupleDict
from .helpers import commas
from .content import TextContent

class Call:

    def __init__(self, function, args: ArgsTupleDict):
        self.function = function
        self.args = args


#
class CallRenderer:

    def __init__(self, *,
                 is_code=True, show_function=True, prefix="", postfix="",
                 max_width=0, css_properties=None):

        self.is_code = is_code
        self.show_function = show_function
        self.prefix = prefix
        self.postfix = postfix
        self.max_width = max_width
        self.css_properties = css_properties if css_properties is not None else {}


    def __repr__(self):
        result =  f"<CallRenderer"
        if self.show_function: result += f" show_function={self.show_function}"
        if self.max_width: result += f" m_w={self.max_width}"
        if self.prefix: result += f" pre={self.prefix}"
        if self.postfix: result += f" post={self.postfix}"
        result += ">"
        return result


    def visible_function_name(self, call):
        if not self.show_function:
            return None
        function_name = self.show_function
        if function_name is True:
            function_name = call.function.__name__
        return function_name


    def render(self, call: Call):
        text = ', '.join(call.args.tokens())
        # show_function can be either a function name to display
        # or just True in which case we compute the name
        # of the call's function attribute
        function_name = self.visible_function_name(call)
        if function_name:
            text = f"{function_name}({text})"
        text = f"{self.prefix}{text}{self.postfix}"
        result = (TextContent(text)
                  .add_css_properties(self.css_properties)
                  .add_css_properties({'align-self': 'center'}))
        if self.max_width:
            result.add_layout({'max-width': f"{self.max_width}em"})
        result.set_is_code(self.is_code)
        return result




class PPrintCallRenderer(CallRenderer):
    def __init__(self, width=80, indent=2, compact=True, *args, **kwds):
        self.width = width
        self.indent = indent
        self.compact = compact
        super().__init__(*args, **kwds)


    def render(self, call: Call):
        # try to render with no width limit, if it fits it's OK
        raw_content = super().render(call)
        simple_case = raw_content.text
        if len(simple_case) <= self.width:
            return raw_content
        # else
        indent = self.indent
        width = self.width
        sep = indent*' '
        html = ""
        html += self.prefix
        function_name = self.visible_function_name(call)
        if function_name:
            html += function_name + "(\n"
        def indent_pformat(pformat_result):
            return sep + pformat_result.replace("\n", "\n"+sep)
        args_tokens = [
            pprint.pformat(arg, width=width-indent, indent=indent)
            for arg in call.args.args]
        keyword_tokens = [
            f"{k}={pprint.pformat(v, width=width-indent, indent=indent)}"
            for (k, v) in call.args.keywords.items()]
        tokens = [indent_pformat(x) for x in args_tokens + keyword_tokens]
        html += (",\n").join(tokens)
        if function_name:
            html += ")\n"
        html += self.postfix
        return (TextContent(html, css_properties=self.css_properties)
                .set_is_code(self.is_code))


class IsliceRenderer(PPrintCallRenderer):

    @staticmethod
    def pretty_slice(generator_args):
        if not generator_args.islice:
            return "all results"
        if len(generator_args.islice) == 1:
            end, = generator_args.islice
            return f"iters →{end}"
        if len(generator_args.islice) == 2:
            beg, end = generator_args.islice
            if beg is not None:
                return f"iters {beg}→{end}"
            else:
                return f"iters →{end}"
        if len(generator_args.islice) == 3:
            beg, end, step = generator_args.islice
            if beg is not None:
                return f"iters {beg}→{end}/{step}"
            else:
                return f"iters →{end}/{step}"
        return f"ERROR: unknown islice {generator_args.islice}"

    # by design this should be called on Call objects
    # whose Args instance is a GeneratorArgs
    def render(self, generator_call):
        generator_args = generator_call.args
        default_content = super().render(generator_call)
        if generator_args.islice is None:
            return default_content
        # xxx not super clean...
        default_content.text += f"\n<b>{self.pretty_slice(generator_args)}</b>"
        return default_content




from .args import Args, GeneratorArgs
from .renderer import Renderer, PPrintRenderer, MultilineRenderer
from .callrenderer import CallRenderer, PPrintCallRenderer, IsliceRenderer

from .exercise_function import ExerciseFunction, ExerciseFunctionNumpy
from .exercise_regexp import ExerciseRegexp, ExerciseRegexpGroups
from .exercise_generator import ExerciseGenerator
from .exercise_class import ExerciseClass, ClassScenario, ClassExpression, ClassStatement

from .content import (TextContent, CodeContent, MathContent,
                      MarkdownContent, MarkdownMathContent)
from .quiz import (Quiz, QuizQuestion, Explanation,
                   Option, CodeOption, MathOption, MarkdownOption, MarkdownMathOption)
from .quiz_loader import run_yaml_quiz

from .quiz_help import quiz_help

from .version import __version__



from nbautoeval import Args, ExerciseClass, ClassScenario, ClassExpression

class Comma:

    def __init__(self, *, a=None, b=None):
        if a is None and b is None:
            raise ValueError(f"a or b must be specified")
        if a is not None:
            self.a = a 
        else:
            self.a = 2 * b
    def __repr__(self):
        return f"{self.a}"

scenario1 = ClassScenario(
    # arguments to the constructor
    Args(a=10),
    # a list of expressions, with 
    # INSTANCE and CLASS replaced as appropriate
    ClassExpression("repr(INSTANCE)"),
)

exo_comma = ExerciseClass (
    Comma, [scenario1],
    layout='pprint')

if __name__ == '__main__':
    exo_comma.correction(Comma)




# an earlier version was relying on markdown2
#from myst_parser.main import to_html, default_parser

from markdown_it import MarkdownIt

from ipywidgets import HTML, HTMLMath, Layout

class Content:

    """
    this class is what a renderer returns
    it models what will ultimately go into a visible Area, that is to say
    . css properties
    . classes
    . a layout (in the ipywidgets sense), here we model this as a dictionary

    the most useful verions of this is TextContent, that has
    . a textual contents
    . whether it is code
    . whether it contains math


    this is materialized by its widget() method that should return an ipywidget
    """

    def __init__(self, *, css_properties=None, classes=None, layout_dict=None):
        self.css_properties = css_properties or {}
        self.classes = classes or []
        self.layout_dict = layout_dict or {}
        self._widget_instance = None

    def __repr__(self):
        return (f"<{type(self).__name__} with {len(self.css_properties)} css props "
                f"{len(self.classes)} classes> "
                f"and {len(self.layout_dict)} items in its ipywidgets layout")

    def style(self):
        return "".join(f"{k}:{v};" for (k, v) in self.css_properties.items())
    def style_attribute(self):
        return f'style="{self.style()}"' if self.css_properties else ''



    def add_css_properties(self, css_properties: dict):
        self.css_properties.update(css_properties)
        return self

    def add_class(self, cls):
        self.classes.append(cls)
        return self

    def add_classes(self, classes):
        self.classes.extend(classes)
        return self

    def add_layout(self, layout_dict):
        self.layout_dict.update(layout_dict)
        return self


    def widget(self):
        """
        this generic code is only about caching
        each subclass should provide their _widget_(self) method
        """
        if self._widget_instance is None:
            self._widget_instance = self._widget_()           # pylint: disable=no-member
        return self._widget_instance



class TextContent(Content):
    '''
    the generic mechanics to turn a text into a widget
    several settings are available, which by default
    are all turned off

    is_code: will cause the whole content to be taken as code
    needs_math: will create a HTMLMath widget instead of plain HTML
    has_markdown: text first goes through a markdown stage
    '''
    def __init__(self, text,
                 is_code=False, needs_math=False, has_markdown=False,
                 **kwds):
        super().__init__(**kwds)
        # make yaml more robust, especially with answers like 'no'
        # that, when not quoted in yaml, produce False that is a bool
        if not isinstance(text, str):
            print(f"WARNING: building a TextContent from non-str {text}")
            text = str(text)
        self.text = text
        self.is_code = is_code
        self.needs_math = needs_math
        self.has_markdown = has_markdown

    # used in particular in quiz's sanity checks
    def __str__(self):
        return self.text

    def set_is_code(self, is_code):
        """
        the set_ methods allow to alter the object with chained statements
        my_content.set_is_code().set_needs_math()
        """
        self.is_code = is_code
        return self

    def set_needs_math(self, needs_math):
        self.needs_math = needs_math
        return self

    def set_has_markdown(self, has_markdown):
        self.has_markdown = has_markdown
        return self

    def _widget_(self):

        text = self.text
        if self.has_markdown:
            parser = MarkdownIt("commonmark")
            text = parser.render(text)
        if self.is_code:
            text = f"<pre>{text}</pre>"
        if self.css_properties:
            text = f"<span {self.style_attribute()}'>{text}</span>"
        widget_class = HTML if not self.needs_math else HTMLMath
        result = widget_class(text, layout=self.layout_dict)
        for cls in self.classes:
            result.add_class(cls)
        return result

# get predefined settings just by the class name
class CodeContent(TextContent):

    def __init__(self, text, **kwds):
        super().__init__(text, **kwds)
        self.is_code = True

class MathContent(TextContent):

    def __init__(self, text, **kwds):
        super().__init__(text, **kwds)
        self.needs_math = True

class MarkdownContent(TextContent):
    def __init__(self, text, **kwds):
        super().__init__(text, **kwds)
        self.has_markdown = True

class MarkdownMathContent(TextContent):
    """
    most useful - most common
    """
    def __init__(self, text, **kwds):
        super().__init__(text, **kwds)
        self.has_markdown = True
        self.needs_math = True


class CssContent(Content):
    """
    dedicated to inject CSS into HTML through
    a <style> tag that has display='none'
    """

    def __init__(self, plain_css, **kwds):
        self.plain_css = plain_css
        super().__init__(**kwds)

    def __repr__(self):
        return f"<CssContent {self.plain_css}>"

    def _widget_(self):
        html = f"<style>{self.plain_css}</style>"
        return HTML(html, layout={'display': 'none'})


class ResultContent(Content):

    def __init__(self, boolean, **kwds):
        self.boolean = boolean
        super().__init__(**kwds)
        self.css_properties['align-self'] = 'center'


    def _widget_(self):
        symbol = "fa-check" if self.boolean else "fa-close"

        html = f'<span {self.style_attribute()} class="fa {symbol}"></span>'
        result = HTML(html)
        result.add_class('result')
        result.add_class(self.the_class())
        for cls in self.classes:
            result.add_class(cls)
        return result


    def the_class(self):
        return 'ok' if self.boolean else 'ko'



from IPython.display import display
from ipywidgets import HTML
from markdown_it import MarkdownIt

french = """
* le quiz est contenu dans **une cellule de code**, 
  que **vous devez évaluer** pour faire **apparaitre les questions**
* toutes les questions ont **au moins une réponse valable**,  
  si vous ne cochez aucune réponse on considère que vous préférez ne pas répondre
* il y a deux sortes de questions,  
  * à **choix unique (sans ♧)** : ne peuvent avoir qu'une seul réponse valable 
  * à **choix multiple (avec ♧)** : celles ou plusieurs cases peuvent être cochées 
* **choix unique** : le barême est indiqué, par exemple `4 pts / -1 pt / 0 pt` signifie
  * 4 points pour une **bonne réponse**
  * -1 point en cas de **réponse fausse**
  * 0 point si vous ne répondez **pas du tout** (à nouveau : si vous ne cochez aucune option)
* **choix multiple** : même principe, mais le barême apparaît comme par exemple 
  `4 pts` / `??` / `-2 pts` / `0 pt`, car **la note est progressive** en fonction
  du nombre de réponses justes; dans ce cas le barême pour toutes les réponses justes 
  ici encore est de 4pts, toutes les réponses fausses vaut -2pts, et 0pt si vous ne répondez pas;  
  si dans cet exemple vous avez 2 réponses justes sur trois, votre note apparaitra comme 
  `4 pts` / **`2 pt`** / `-2 pts` / `0 pt`; et vous avez 2pts parce vous obtenez
  les deux tiers entre le min et le max des notes; 
* vous avez un nombre fixe d'essais pour répondre à un quiz - indiqué dans le bouton de soumission;  
  une fois votre nombre d'essais épuisé - ou plus tôt si vous avez tout juste -
  la correction vous montre (en rouge vif) les cases où vous vous êtes trompé ;
  les cases avec un ☛ donnent des explications supplémentaires
* le résultat final correspond à votre meilleur essai;
  les essais sont notés indépendamment les uns des autres,
   il n'y a pas d'effet cumulatif : 
   si par exemple vous obtenez 10/20 au premier essai en ayant juste à la question 1,
   et 12/20 au deuxième essai en ayant faux à la question 1,
   on retient le 12/20, la bonne réponse à l'essai #1 n'affecte pas l'essai #2
"""
  
english = """
* le quiz is contained in a **code cell**, most often empty at first, so **you need 
  to evaluate it** to get the questions to show up
* all questions have at least one valid answer;  
  if you do not tick any box in a question, it means you prefer not to answer
* there are 2 kinds of questions,  
  * **single choice (without ♧)** : can only have one valid answer
  * **multiple multiple (qith ♧)** : several boxes may be ticked
* **single choice** : the score is shown like e.g. `4 pts / -1 pt / 0 pt`, meaning 
  * 4 points for a **valid answer**
  * -1 point for a **wrong answer**
  * 0 point if you do not answer (again, if you do not tick any box)
* **multiple choice** : same idea, but the score will show like e.g. 
  `4 pts` / `??` / `-2 pts` / `0 pt`,  
  because **the grade is progressive** according 
  to the number of right answers;   
  in that example, again you'll get 4pt for all answers correct, 
  -2pts for all answers wrong, and 0pt if you prefer to opt out;  
  in this example if you have 2 correct answers out of 3, you will see  
  `4 pts` / **`2 pt`** / `-2 pts` / `0 pt`; and you get 2pts because 
  you obtain 2/3 between the min and max grades
* you have a fixed number of attempts to answer a quiz - shown in the submit button;
  once that number has run out - or earier if you get the max score -
  the correction outlines (in stark red) the boxes that you have gotten wrong;
  additional areas with a ☛ may provide additional explanation
* your final score corresponds to your best attempt; 
  attempts are independent from one another, there is no cumulative effect :  
  if for example you get 10/20 on your first attempt, with question 1 being right,
  and 12/20 on your second one with question 1 begin wrong, the second score is retained,
  the correct answer on attempt #1 does not impact attempt #2
"""



langs = {'fr': french, 'en': english}

def quiz_help(lang='en'):
    markdown = langs.get(lang, f"*no available help for lang {lang}")
    html = MarkdownIt("commonmark").render(markdown)
    display(HTML(html))



# -*- coding: utf-8 -*-

# pylint: disable=c0111, w0703


# for logging
import os
import time
import json

from pathlib import Path

########## logging
def _log_path():
    """
    returns a Path object where to log attempts
    
    can be specified with env. variable NBAUTOEVAL_LOG
    defaults to $HOME/.nbautoeval
    """
    # use the env variable if set, or fallback to default
    from_env = os.environ.get('NBAUTOEVAL_LOG')
    
    result = Path(from_env) if from_env else Path.home() / ".nbautoeval"
    # we need to return a file and to make sure its dir. exists
    if result.is_dir():
        return result / ".nbautoeval"
    elif result.parent.is_dir():
        return result
    else:
        result.parent.mkdir(parents=True)
        return result

def _log2_path():
    return _log_path().with_suffix(".trace")

def _storage_path():
    return _log_path().with_suffix(".storage")

# historically for each attempt we stored in 
#    ~/.nbautoeval 
# a line with
# <date> <userid> <exoname> <details...>
# with details depending on the kind of exercise 
# * functions get a OK/KO, while
# * quizzes get details on the grade
# because this format has grown fragile it is recommended to read
#   ~/.nbautoeval.trace instead,
# which has the same info but in JSON format
# now, for compatibility we still store the old format
# but the way to compute this <userid> thingy is fragile too
# and it is no longer exposed in the json output anyway
# so from 1.1.6 we just write 'unknown-user' instead of userid

# ----------
# first-generation logging - for exercises initially
def _log_line(exoname, message):
    """
    write a one-liner in the log file that contains
    timestamp unix-uid exoname <some message>
    """
    try:
        now = time.strftime("%D-%H:%M", time.localtime())
        uid = 'unknown-user'
        with _log_path().open('a') as log:
            line = f"{now} {uid} {exoname} {message}\n"
            log.write(line)
    except Exception as exc:
        print(f"nbautoeval could not _log_line {message} - {type(exc)} {exc}")


# ----------
def log_correction(exoname, success):
    """
    this is adapted to binary feedback (good or bad)
    """
    _log_line(exoname, "OK" if success else "KO")
    

def log_quiz(exoname, score, max_score, normalized_score, normalized_max_score):
    """
    this is adapted to quiz-like exercices that 
    come out with a score like 14 / 20
    """
    _log_line(exoname, (f"{score}/{max_score} "
                        f"{normalized_score:.2f}/{normalized_max_score:.2f}"))

# ----------
# second-generation logging - one json record per line
# for a more flexible format
# for quizzes right now, would make sense to extend to exercises as well
# drop uid in the mix, it's just noise, this goes in $HOME already
def _log2_struct(exoname, type, struct: dict):
    # keep stuff in order
    stored = dict(
        exoname=exoname,
        type=type,
        time=time.strftime("%D-%H:%M:%S", time.localtime()))
    stored.update(struct)
    try:
        with _log2_path().open('a') as log:
            print(json.dumps(stored), file=log)
    except Exception as exc:
        print(f"nbautoeval could not _log_struct {struct} - {type(exc)} {exc}")


def log2_correction(exoname, *, success, **optional):
    """
    extensible json-based log for exercises
    """
    struct = dict(success=success)
    struct.update(optional)
    _log2_struct(exoname, 'exercise', struct)


def log2_quiz(exoname, *, attempt, max_attempts, score, max_score, **optional):
    """
    extensible json-based log about a quiz 
    with thorough details on that attempt
    """
    struct = dict(attempt=attempt, max_attempts=max_attempts,
                  score=score, max_score=max_score)
    struct.update(optional)
    _log2_struct(exoname, 'quiz', struct)

# ----------
def storage_read(exoname, attribute, default):
    try:
        with _storage_path().open() as feed:
            storage = json.loads(feed.read())
            return storage[exoname].get(attribute, default)
    except:
        return default
        
def storage_save(exoname, attribute, value):
    try:
        with _storage_path().open() as feed:
            storage = json.loads(feed.read())
    except:
        storage = {}
    
    if exoname not in storage:
        storage[exoname] = {}
    storage[exoname][attribute] = value
    
    with _storage_path().open('w') as writer:
        writer.write(json.dumps(storage))
        

def storage_clear(exoname):
    try:
        with _storage_path().open() as feed:
            storage = json.loads(feed.read())
    except:
        return
    if exoname not in storage:
        return

    del storage[exoname]
    
    with _storage_path().open('w') as writer:
        writer.write(json.dumps(storage))
        
    



import itertools

from nbautoeval import ExerciseGenerator, GeneratorArgs, PPrintRenderer

def differential(iterator):
    previous = next(iterator)
    while True:
        current = next(iterator)
        yield current - previous
        previous = current


def squares():
    return (i**2 for i in itertools.count())

differential_args = [
    GeneratorArgs(itertools.count(), islice=(10,)),
    GeneratorArgs(squares(), islice=(10,)),
]

exo_differential = ExerciseGenerator(
    differential, differential_args,
    copy_mode='tee',  # this is the broken piece in 0.6.1
    max_iterations=20,
    result_renderer=PPrintRenderer(width=20),
    
)



# -*- coding: utf-8 -*-

# pylint: disable=c0111, c0103, r1705, w0703

from ipywidgets import GridBox, Layout

from .args import Args
from .content import TextContent, CssContent, ResultContent
from .callrenderer import Call, CallRenderer
from .renderer import Renderer
from .helpers import default_font_size, default_header_font_size
from .storage import log_correction, log2_correction


DEBUG = False
#DEBUG = True


default_column_headers = (
    "scénario {n}", 
    "appel",
    "attendu",
    "obtenu",
)
# same order
column_span_classes = (
    ["span-1-to-4", "scenario"], [], [], ["span-3-to-4"],
)
# simpler in example mode
example_column_span_classes = (
    ["span-1-to-2", "scenario"], [], [],
)

###
CSS = """
.nbae-cls {
    width: max-content;
/*    border-top: .1em solid black;  */
    border-bottom: .1em solid black;
    padding: 0.1em 0em;
    row-gap: 0.4em;
    column-gap: .4em;
}
.nbae-cls .header {
    padding: 0em 0em 0.3em 0em;
    background-color: #f1fcfc;
    border-bottom: .1em solid #888;
    font-weight: bold;
    border-top-left-radius: .4em;
    border-top-right-radius: .4em;
}
.nbae-cls .header > .widget-html-content {
    justify-content: center;
}

.nbae-cls .scenario.header {
    padding: 0.3em 0em 0em 0em; 
    border-bottom: none;
    border-top: .1em solid #444;
}
.nbae-cls .span-1-to-4 {
    grid-column: 1 / span 4;
}
.nbae-cls .span-3-to-4 {
    grid-column: 3 / span 2;
}
.nbae-cls .span-2-to-3 {
    grid-column: 2 / span 2;
}
.nbae-cls .span-1-to-2 {
    grid-column: 1 / span 2;
}
div.nbae-cls div.widget-html-content pre {
    padding: 0px;
    line-height: 1.15;
}
.nbae-cls .ok, .nbae-cls .ok code {
    background-color: #d4f8e8;
} 
.nbae-cls .ok.even, .nbae-cls .ok code.even {
    background-color: #c4f8d8;
}
.nbae-cls .ko, .nbae-cls .ko code {
    background-color: #ffc6c9;
}
.nbae-cls .ko.even, .nbae-cls .ko code.even {
    background-color: #ffb6b9;
}
.nbae-cls .cell {
    padding: .2em .4em;
    border-radius: .4em;
}
.nbae-cls .cell.example {
    background-color: #f8f8f8;
    padding: .5em 1em;
}
.nbae-cls .cell.example.even {
    background-color: #e8e8e8;
    padding: .5em 1em;
}
.nbae-cls .result {
    padding: 0em .6em 0em .6em;
}
.nbae-cls .ko.result {
    font-weight: bold;
    font-size: '120%';
}
.nbae-cls div.widget-html-content {
    display: flex;
}
"""


class ClassExpression:
    """
    this utility class is used to model a step in a class scenario
    it is basically built from a string where
        'INSTANCE' is the object created in the first step, and
        'CLASS' is the class object itself
    """
    
    def __init__(self, code, statement=False):
        self.code = code
        self.statement = statement
        
    def __repr__(self):
        #  return f"<{type(self).__name__} {self.statement=} {self.code=}>"
        return f"<{type(self).__name__} statement={self.statement} code={self.code}>"
        
        
    def replace(self, varname, classname):
        return self.code.replace("INSTANCE", varname).replace("CLASS", classname)


class ClassStatement(ClassExpression):
    """
    a shortcut to create statements
    """
    def __init__(self, code):
        ClassExpression.__init__(self, code, True)

##########
class ClassScenario:
    """
    Describes a scenario that can be applied to a class

    Typically we want to create an instance (using some args),
    and then run some methods (still with some args)

    So a scenario is defined from
      * one Args instance that is passed to the constructor
      * a list of ClassExpression (or mere str) objects
        
    Example:
      for a Polynom class created from a set of coefficients
      ClassScenario(
          Args(1, 2, 3),
          ClassExpression("INSTANCE"),
          ClassExpression("INSTANCE.derivative()"),
          ClassExpression("INSTANCE + CLASS(3, 4, 5)"),
      )
      
      the expressions will be evaluated for 
      both the reference class and the student's class
      and the results compared with == 
      (unless the validate method is redefined on the Exercise class)
    """

    def __init__(self, init_args, *expressions):
        self.init_args = init_args
        def step(exp):
            return ClassExpression(exp) if isinstance(exp, str) else exp
        self.steps = [step(exp) for exp in expressions]
        if not isinstance(self.init_args, Args):
            print(f"ERROR ClassScenario first parameter needs to be an Args instance")
        for index, step in enumerate(self.steps, 1):
            if not isinstance(step, ClassExpression):
                print(f"ERROR ClassScenario, step #{index}, needs to be a ClassExpression instance")


##########
class ExerciseClass:                                    # pylint: disable=r0902
    """
    Much like the ExerciseFunction class, this allows to define
    an exercise as
    (*) a solution which is a Python class that provides the correct implementation
    (*) a list of scenarios that will be executed on that class

    From that plus a few accessories for fine-grained customization
    we can generate online example and correction.
    
    the check_init flag mostly targets very simple and early exos,
    given to students who do not yet know how to write their own repr();
    of course this assumes to not use a statement in the scenario, as that
    would still trigger repr()
    """

    def __init__(self, solution, scenarios,                     # pylint: disable=r0913
                 *,
                 copy_mode='deep',
                 nb_examples=1,
                 # how to render
                 result_renderer=None,
                 #
                 obj_name='X',
                 #
                 column_headers=None,
                 font_size=default_font_size,
                 header_font_size=default_header_font_size,
                 check_init=True,
                 ):
        # the 'official' solution
        self.solution = solution
        # the inputs - actually Scenario instances
        self.scenarios = scenarios
        # how to copy args
        self.copy_mode = copy_mode
        # how many examples
        self.nb_examples = nb_examples
        # a - preferrably uppercase - string to refer to 
        # current object in the scenarios
        self.obj_name = obj_name
        # 
        self.result_renderer = result_renderer or Renderer()
        # header names 
        self.column_headers = column_headers or default_column_headers
        #
        self.header_font_size = header_font_size
        self.font_size = font_size
        # see above
        self.check_init = check_init
        # computed
        self.name = solution.__name__
        

    def correction(self, stu_class, print_exceptions=False): # pylint: disable=r0914, r0915

        overall = True
        ref_class = self.solution

        headers_props = {'font-size': self.header_font_size}
        body_props = {'font-size': self.font_size}
        contents = []

        for index, scenario in enumerate(self.scenarios, 1):

            classes = ['cell']

            init_args = scenario.init_args
                                 

            # header for scenario
            contents += [TextContent(x.format(n=index))
                         .add_css_properties(headers_props)
                         .add_class('header')
                         .add_classes(span_classes)
                        for (x, span_classes) in zip(self.column_headers, 
                                                     column_span_classes)]
        
            call_renderer = CallRenderer(show_function=self.name, 
                                         prefix=f"{self.obj_name} = ",
                                         postfix=f"; repr({self.obj_name})") 
            init_rendered = call_renderer.render(Call(None, init_args))
            # clone args for both usages
            ref_args = init_args.clone(self.copy_mode)
            stu_args = init_args.clone(self.copy_mode)
                
            # initialize both objects
            try:
                # initialize both objects
                REF = ref_args.init_obj(ref_class)  
                STU = stu_args.init_obj(stu_class)
                
                if not self.check_init:
                    ref_repr = stu_repr = '--unchecked--'
                    is_ok = True
                else:
                    ref_repr, stu_repr = repr(REF), repr(STU)
                    is_ok = self.validate(REF, STU, ref_class, stu_class)
                    if not is_ok:
                        overall = False
                # render that run
                result_content = ResultContent(is_ok)

                contents.append(init_rendered
                                .add_classes(classes)
                                .add_css_properties(body_props))
                contents.append(self.result_renderer.render(ref_repr)
                                .add_classes(classes).add_class('ok')
                                .add_css_properties(body_props))
                contents.append(self.result_renderer.render(stu_repr)
                                .add_classes(classes)
                                .add_class(result_content.the_class())
                                .add_css_properties(body_props))
                contents.append(result_content
                                .add_classes(classes)
                                .add_css_properties(body_props))
                
            except Exception as exc:
                if print_exceptions:
                    import traceback
                    traceback.print_exc()
                error = f"Exception {type(exc)} {exc}"
                contents.append(init_rendered
                                .add_classes(classes)
                                .add_css_properties(body_props))
                contents.append(TextContent(error)
                                .add_classes(classes)
                                .add_class('ko')
                                .add_class('span-2-to-3')
                                .add_css_properties(body_props))
                contents.append(ResultContent(False)
                                .add_css_properties(body_props))
                overall = False
                continue

            # other steps of that scenario; first step was __init__
            for step_index, step in enumerate(scenario.steps, 2):
                code_ref = step.replace("REF", "ref_class")
                code_stu = step.replace("STU", "stu_class")
                display = step.replace(self.obj_name, ref_class.__name__)
                if step.statement:
                    display += f"; {self.obj_name}"

                classes = ['cell']
                if step_index % 2 == 0:
                    classes.append('even')
                # the function to use to run code, whether it's a statement
                # or an expression; in the former case of course, there is no need 
                # to cmpare results as they are None, but that's not important
                code_runner = exec if step.statement else eval
                ref_result = code_runner(code_ref)
                try:
                    stu_result = code_runner(code_stu)
                    if step.statement:
                        stu_result = repr(STU)
                        ref_result = repr(REF)
                    is_ok = self.validate(ref_result, stu_result, ref_class, stu_class)
                    if not is_ok:
                        overall = False
                    result_content = ResultContent(is_ok)

                except Exception as exc:
                    result_content = ResultContent(False)
                    overall = False
                    stu_result = f"Exception {type(exc)}: {exc}"
                    if step.statement:
                        ref_result = repr(REF)

                contents.append(TextContent(display)
                                .set_is_code(True)
                                .add_classes(classes)
                                .add_css_properties(body_props))
                contents.append(self.result_renderer.render(ref_result)
                                .add_classes(classes).add_class('ok')
                                .add_css_properties(body_props))
                contents.append(self.result_renderer.render(stu_result)
                                .add_classes(classes)
                                .add_class(result_content.the_class())
                                .add_css_properties(body_props))
                contents.append(result_content
                                .add_classes(classes)
                                .add_css_properties(body_props))

        log_correction(self.name, overall)
        # xxx would make sense to expose how many examples were right or wrong
        log2_correction(self.name, success=overall)

        contents.append(CssContent(CSS))

        gridbox_layout  = Layout(grid_template_columns = 'max-content 1fr 1fr max-content',
                                 max_width="100%")
        grid = GridBox([content.widget() for content in contents],
                       layout=gridbox_layout).add_class("nbae-cls")
        return grid


    def example(self):                                  # pylint: disable=r0914
        """
        display a table with example scenarios
        """
        ref_class = self.solution

        headers_props = {'font-size': self.header_font_size}
        body_props = {'font-size': self.font_size}
        contents = []

        how_many_samples = self.nb_examples if self.nb_examples \
                           else len(self.scenarios)

        sample_scenarios = self.scenarios[:how_many_samples]
        for index, scenario in enumerate(sample_scenarios, 1):

            classes = ['cell', 'example']
            # start of scenario
            contents += [
                TextContent(x.format(n=index))
                .add_css_properties(headers_props)
                .add_class('header')
                .add_classes(span_classes)
                for (x, span_classes) in zip(self.column_headers,
                                             example_column_span_classes)]

            # object construction
            init_args = scenario.init_args.clone(self.copy_mode)
            SAMPLE = init_args.init_obj(ref_class)      # pylint: disable=unused-variable

            call_renderer = CallRenderer(show_function=self.name, 
                                         prefix=f"{self.obj_name} = ",
                                         postfix=f"; repr({self.obj_name})") 
            init_rendered = call_renderer.render(Call(None, init_args))
            contents.append(init_rendered
                            .add_classes(classes)
                            .add_css_properties(body_props))
            contents.append(self.result_renderer.render(repr(SAMPLE))
                            .add_classes(classes)
                            .add_css_properties(body_props))

            for step_index, step in enumerate(scenario.steps, 2):
                display = step.replace(self.obj_name, ref_class.__name__)
                code = step.replace("SAMPLE", "ref_class")
                code_runner = exec if step.statement else eval
                ref_result = code_runner(code)
                classes = ['cell', 'example']
                if step_index % 2 == 0:
                    classes.append('even')
                
                ### display
                if step.statement:
                    display += f"; repr({self.obj_name})"
                    ref_result = repr(SAMPLE)
                contents.append(TextContent(display)
                                .set_is_code(True)
                                .add_classes(classes)
                                .add_css_properties(body_props))
                contents.append(self.result_renderer.render(ref_result)
                                .add_classes(classes)
                                .add_css_properties(body_props))

        contents.append(CssContent(CSS))

        gridbox_layout  = Layout(grid_template_columns = 'max-content max-content')
        grid = GridBox([area.widget() for area in contents],
                       layout=gridbox_layout).add_class("nbae-cls")
        return grid
    

    def validate(self, ref_res, stu_res, ref_class, stu_class):
        """
        how to compare the results as obtained from
        * the solution instance
        * the student instance
        * the solution class
        * the student class 
        
        it is not possible in the general case to use == on objects
        since the 2 classes are different, and so == will always return False
        """
        if not isinstance(ref_res, ref_class) and not isinstance(stu_res, stu_class):
            # assuming both are regular Python objects 
            # use ==
            return ref_res == stu_res
        if isinstance(ref_res, ref_class) and isinstance(stu_res, stu_class):
            # compare repr's
            return repr(ref_res) == repr(stu_res)
        else:
            # something is wrong
            return False



# -*- coding: utf-8 -*-
from nbautoeval import ExerciseFunction, Args, CallRenderer

#
# example how to use
# 

# @BEG@ week=0 sequence=0 name=curve
def curve(a, b, c=12):
    return a ** 2 + 3 * a * b + c
# @END@

# it is unwise to have one dataset is shared between 2 exercises
# so let's create one for each exercise
def inputs_curve ():
    return [
        Args(1, 2),
        Args(1, 2, 5),
        Args(1, 1),
        Args(1, 5),
        Args(2, 1, 10),
        Args(2, 2),
        Args(2, 3, 10),
        Args(4, 4),
    ]

exo_curve = ExerciseFunction(curve, inputs_curve())

exo_curve_noname = ExerciseFunction(
    curve, inputs_curve(), 
    call_renderer=CallRenderer(show_function=False))



# 
# a very small class with a property
# 

MIN = 0
MAX = 100

class Gauge:
    
    def __init__(self, x):
        self.x = x
        
    def _get_x(self):
        return self._x
    
    def _set_x(self, x):
        self._x = min(MAX, max(MIN, x))
        
    x = property(_get_x, _set_x)
    
    def __repr__(self):
        return f"{self._x}"


# create an exercise instance

from nbautoeval.exercise_class import (
    ExerciseClass, ClassScenario, ClassExpression, ClassStatement)
from nbautoeval.args import Args

scenario1 = ClassScenario(
    # arguments to the constructor
    Args(10),
    # statements need to be tagged as such 
    ClassStatement("INSTANCE.x = 50"),
    ClassStatement("INSTANCE.x = 2000"),
)

scenario2 = ClassScenario(
    # arguments to the constructor
    Args(1000),
    # note that a str object passed here is actually
    # used to create a ClassExpression object
    ClassStatement("INSTANCE.x = -1000"),
)


exo_gauge = ExerciseClass(
    Gauge, [scenario1, scenario2],
    obj_name = "gauge",
    )

if __name__ == '__main__':
    exo_gauge.example()
    exo_gauge.correction(Gauge)




#!/usr/bin/env python3

from pathlib import Path

import setuptools


def contents(localfile):
    with (Path(__file__).parent / localfile).open() as f:
        return f.read()

# https://packaging.python.org/guides/single-sourcing-package-version/
# set __version__ by read & exec of the python code
# this is better than an import that would otherwise try to
# import the whole package, and fail if a required module is not yet there
VERSION_FILE = Path(__file__).parent / "nbautoeval" / "version.py"
ENV = {}
with VERSION_FILE.open() as f:
    exec(f.read(), ENV)                                 # pylint: disable=w0122
__version__ = ENV['__version__']

setuptools.setup(
    name             = "nbautoeval",
    version          = __version__,
    author           = "Thierry Parmentelat",
    author_email     = "thierry.parmentelat@inria.fr",
    description      = "A mini framework to implement auto-evaluated exercises in Jupyter notebooks",
    long_description = contents("README.md"),
    long_description_content_type = "text/markdown",
    license          = "CC BY-SA 4.0",
    keywords         = "jupyter notebooks exercises",
    url              = "https://github.com/parmentelat/nbautoeval",
    packages         = setuptools.find_packages(),
    install_requires = [
        'ipython',
        'ipywidgets',
        'numpy', 
        'PyYAML',
        'myst_parser',
    ],
    classifiers      = [
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Programming Language :: Python",
        # pypi won't let then in
        # "Framework :: Jupyter",
    ],
)



# -*- coding: utf-8 -*-

# pylint: disable=c0111, r1705

import copy
import pprint
import itertools
from collections.abc import Iterable, Iterator

from .helpers import custom_repr

####################
# From June 2016, this class should not need to be used directly
# as Args would allow to build it with a nicer interface
class ArgsTupleDict:
    """
    The most general form of a function argument list is made
    of a tuple and a dictionary

    Example:
    my_input = ArgsTupleDict((1, 2), {'a': []})
    my_input.call(foo)
    would then return the result of
    foo(1, 2, a=[])
    """
    def __init__(self, _args=None, _keywords=None):
        # expecting a tuple or a list
        self.args = _args if _args is not None else tuple()
        # expecting a dictionary
        self.keywords = _keywords if _keywords is not None else {}
        # can be overridden later on using 'render_function_name'
        self.function_name = None
        # can be overridden later on using 'render_prefix'
        self.prefix = ""
        self.postfix = ""


    def __repr__(self):
        result = f"<Args {self.prefix}{self.function_name}{self.args}{self.postfix} "
        if self.keywords:
            result += ("Keywords:" + 
                       ",".join(f"{k}={v}" for (k, v) in self.keywords.items()))
        result += ">"
        return result

    def call(self, function, debug=False):
        if debug:
            print(f"calling {function.__name__} *{self.args} **{self.keywords}")
        return function(*self.args, **self.keywords)

    def init_obj(self, klass, debug=False):
        if debug:
            print(f"creating object in class {klass.__name__}, *{self.args} **{self.keywords}")
        return klass(*self.args, **self.keywords)

    def call_obj(self, obj, methodname, debug=False):
        if debug:
            print(f"calling method {methodname} on object {obj} *{self.args} **{self.keywords}")
        method = getattr(obj, methodname)
        return method(*self.args, **self.keywords)

    def clone(self, copy_mode):
        "clone this input for safety"
        if copy_mode == 'shallow':
            return copy.copy(self)
        elif copy_mode == 'deep':
            return copy.deepcopy(self)
        else:
            return self
    
    # # rendering  
    def tokens(self):
        """
        returns a list of strings
        e.g. 
        Args(1, 2, x=1).tokens -> ['1', '2', 'x=1']
        """
        return ( [custom_repr(x) for x in self.args] 
                +[f"{k}={custom_repr(v)}" for (k, v) in self.keywords.items()])
        

# simplified for when no keywords are required
class Args(ArgsTupleDict):
    """
    In most cases though, we do not use keywords so it is more convenient to
    just pass a list of arguments

    Example:
    my_input = Args (1, 2, 3)
    my_input.call(foo)
    would then return the result of
    foo(1, 2, 3)

    """
    def __init__(self, *args, **kwds):
        # it is NOT *args here, this is intentional
        ArgsTupleDict.__init__(self, args, kwds)


class GeneratorArgs(Args):
    """
    GenArgs is like Args but designed for use 
    with an ExerciseGenerator
    
    See exercises/squares.py for an example of how to use it
    """
    def __init__(self, *args, islice=None, **kwds):
        super().__init__(*args, **kwds)
        self.islice = islice


    def call(self, function, debug=False):
        # call() already has it's contents converted
        # to a list because comparison of results alreqdy has taken place
        iterable = super().call(function)
        if not isinstance(iterable, Iterable):
            raise TypeError(f"not an iterable! received a {type(iterable).__name__} instance: {iterable}")
        if not self.islice:
            return iterable
        return list(itertools.islice(iterable, *self.islice))


    # handles iterator when part of self.args
    # this is a part of aborted 0.6.1
    # intention was to be smart about copying iterators
    # but that did not work out too well
    # kept this code for future reference only
    # but it not used be default with 0.6.2 version
    def copy_for_tee(self, copy_mode):
        copy1, copy2 = type(self)(), type(self)()
        for att in ('keywords', 'function_name'):
            setattr(copy1, att, copy.copy(getattr(self, att)))
            setattr(copy2, att, copy.copy(getattr(self, att)))
        a1, a2 = [], [] 
        for arg in self.args:
            if not isinstance(arg, Iterator):
                a1.append(copy.copy(arg))
                a2.append(copy.copy(arg))
            else:
                c1, c2 = itertools.tee(arg, 2)
                a1.append(c1)
                a2.append(c2)
        copy1.args = tuple(a1)
        copy2.args = tuple(a2)
        return copy1, copy2
    


from itertools import count

from nbautoeval import ExerciseGenerator, GeneratorArgs

def primes():
    cache_primes = set()
    
    # print(f"primes()")
    for n in count(2):
        # print(f"n={n}")
        for i in range(2, n):
            if i not in cache_primes:
                continue
            if n % i == 0:
                break
        else:
            cache_primes.add(n)
            yield n
            

primes_inputs = [
    GeneratorArgs(islice=(20,)),
    GeneratorArgs(islice=(None, 20, 2)),
    GeneratorArgs(islice=(100, 101)),
    GeneratorArgs(islice=(101, 102)),
    GeneratorArgs(islice=(1001, 1002)),
]

# max_iterations is mostly a provision to avoid endless loops
# as of 0.6 it seems like setting it to a big value like 10_000
# causes trouble, so ..
exo_primes = ExerciseGenerator(
    primes, primes_inputs, max_iterations=101,
    nb_examples=0,
)

exo_primes_no_limit = ExerciseGenerator(
    primes, primes_inputs, 
    nb_examples=0,
)


from pathlib import Path

# pip install PyYAML
import base64
import binascii
import yaml

# need to import them all here even when not explicitly used
# so that yaml code can refer to these classes
from .content import (TextContent, CodeContent, MathContent,
                      MarkdownContent, MarkdownMathContent)
from .quiz import (Quiz, QuizQuestion, Explanation,
                   Option, CodeOption, MathOption, MarkdownOption, MarkdownMathOption,
                   DEFAULT_OPTION_CLASS)


from .quiz import DEFAULT_CONTENT_CLASS

# quite straightforward, we build Python objects
# from a YAML regular Python object
# the purpose of object_from_dict to build
# is to create Python objects from the YAML (dict) data
# so the `type` field allows to chose the Python class name
#
#
# for more convenience some components in the Quiz structure
# may be entered as either a plain str, or a Content object
# example
# the-quiz:
#   type: Quiz
#   # one can either do
#   explanation: some text right here
#   # // or //
#   explanation:
#     type: MarkdownMathContent
#     text: |
#       here you have specified the content type
#
# this goes for
# . QuizQuestion.question            - the main question asked
# . QuizQuestion.question_sequel     - optional question follow-up
# . quizQuestion.option_none         - optional option object for 'none of the others'
# . QuizQuestion.explanation         - optional related explanation
# . all the QuizQuestion.option's              - the main text of the option
#   all the QuizQuestion.option.explanation's  - optional related explanation

class YamlLoader:

    """
    allows to parse a YAML file
    do some high-levels checks
    and build a Quiz object given its name
    """

    def __init__(self, filename_or_path):
        self.path = Path(filename_or_path)
        with self.path.open('rb') as feed:
            binary_content = feed.read()
        # first try, yaml-read as-is
        try:
            self.raw = yaml.safe_load(binary_content.decode(encoding="utf-8"))
            # somehow the exception does not always trigger
            # and in that case we receive a str
            if not isinstance(self.raw, dict):
                raise binascii.Error(None)
        except binascii.Error:
            self.raw = yaml.safe_load(base64.standard_b64decode(binary_content))
        except Exception:
            print(f"Cannot load quiz from yaml {self.path}")
            raise

    def iterate_on(self, typename):
        for name, item in self.raw.items():
            if 'type' in item and item['type'] == typename:
                yield name, item

    def check(self, boolean, message):
        if not boolean:
            raise ValueError(message)

    def rain_check(self):
        # at least one of each
        assert(list(self.iterate_on('Quiz')))
        questions = list(self.iterate_on('QuizQuestion'))
        assert(questions)

        for quizname, quiz in self.iterate_on('Quiz'):
            self.check('questions' in quiz, "Quiz must have questions")
            if isinstance(quiz['questions'], str):
                quiz['questions'] = quiz['questions'].split()
            for qname in quiz['questions']:
                if qname not in self.raw:
                    raise ValueError(
                        f"question named `{qname}` is used in "
                        f"quiz {quizname} "
                        f"but not found in YAML")

        # xxx
        # check for the class names used in the various _type fields

    # build (real instance) object from the
    # dictionary that comes out of yaml
    # opt_type_field_name is meant to allow the yaml desc. to
    # chose another class than the default
    # use opt_type_field_name=None if that's not a flexible option
    def object_from_dict(self, yaml_dict, default_class, opt_type_field_name):
        if not isinstance(yaml_dict, dict):
            print(f"WARNING: expecting a dict, got a {type(yaml_dict).__name__} "
                  f"object {yaml_dict} instead")
        args = yaml_dict.copy()
        cls = default_class
        if opt_type_field_name in args:
            # locate class
            cls = globals()[args[opt_type_field_name]]
            del args[opt_type_field_name]
        # the 'type' attribute is to be removed throughout
        # not only when opt_type_field_name == 'type'
        if 'type' in args:
            del args['type']
        # troubleshoot - we have no line number to offer, so
        # try to be helpful so users can pinpoint their mistake
        if self.debug:
            if cls.__name__ == 'QuizQuestion':
                print(f"building a {cls.__name__} : "
                      f"question={args['question'].text[:20]}...")
            elif 'Option' in cls.__name__:
                print(f"building a {cls.__name__} : "
                      f"text={args['text'][:20]}...")
            elif self.debug == 2:
                print(f"building a {cls.__name__} from "
                      f"attributes {list(args.keys())}")
        # what remains is passed as parameters to the constructor
        return cls(**args)

    def flexible_object(self, str_or_dict, cls):
        if isinstance(str_or_dict, str):
            return cls(str_or_dict)
        elif isinstance(str_or_dict, dict):
            return self.object_from_dict(str_or_dict, cls, "type")
        else:
            raise TypeError(f"expecting str or dict, "
                            f"got {type(str_or_dict.__name__)} {str_or_dict}")

    # recursively browse a yaml structure to apply the transformation to objects
    # logic is depth-first scan, and when coming back up apply any additional
    # change
    def process(self, yaml_dict):
        if not isinstance(yaml_dict, dict):
            return yaml_dict
        for k, v in yaml_dict.items():
            if isinstance(v, dict):
                self.process(v)
            elif isinstance(v, list):
                yaml_dict[k] = [self.process(x) for x in v]
            if k == 'explanation':
                # try this
                explanation = self.flexible_object(v, Explanation)
                # but if there was a type specified we still need the Explanation
                if not isinstance(explanation, Explanation):
                    explanation = Explanation(explanation)
                yaml_dict[k] = explanation
            elif k in ('question', 'question_sequel'):
                yaml_dict[k] = self.flexible_object(v, DEFAULT_CONTENT_CLASS)
            elif k in ('option_none'):
                yaml_dict[k] = self.flexible_object(v, DEFAULT_OPTION_CLASS)
            elif k in ('options'):
                yaml_dict[k] = [
                    self.flexible_object(opt, DEFAULT_OPTION_CLASS)
                    for opt in v
                ]
            elif k in ('questions'):
                yaml_dict[k] = [
                    self.object_from_dict(self.process(q), QuizQuestion, 'type')
                    for q in v
                ]
            elif isinstance(v, dict) and 'type' in v:
                yaml_dict[k] = self.object_from_dict(v, None, 'type')
        return yaml_dict


    def build_quiz(self, exoname, debug):
        import pdb
        # pdb.set_trace()
        self.debug = debug
        # first make sure the exoname makes sense
        self.check(exoname in self.raw, f"quiz {exoname} not found")
        yaml_quiz = self.raw[exoname]

        # stage1 on questions: build QuizQuestion instances
        processed_questions = {
            name: self.object_from_dict(self.process(yaml_question), QuizQuestion, None)
            for (name, yaml_question) in self.iterate_on('QuizQuestion')
        }
        """
        for qname, question in processed_questions.items():
            print(f"question {qname} -> {question}")
            break
        """

        # artificially inject exoname from YAML id
        if 'exoname' not in yaml_quiz:
            yaml_quiz['exoname'] = exoname
        quiz = self.object_from_dict(
            # quiz_type not yet used but well
            yaml_quiz, Quiz, "type")
        quiz.questions = [
            processed_questions[qname] for qname in quiz.questions
        ]

        return quiz

#
# a little salt to make it not completey obvious
# how to cheat by looking into the quiz source
# typically using $HOME/.yaml to store source might be
# a good idea
# typically search {radical}, {radical}.yaml {radical}.yml
# in the following dirs (if under HOME)
# .  ./yaml  ./.yaml
# .. ../yaml ../.yaml
# ../.. ../../yaml ../../.yaml
# HOME HOME/yaml HOME/.yaml
#
def locate_from_radical(filename_or_path, debug):

    # skip heuristics if its an existing filename
    checking = Path(filename_or_path)
    if checking.exists():
        return checking
    if checking.is_absolute():
        if debug:
            print(f"cannot use absolute but non existing path {filename_or_path}")
        return None

    # if provided a relative path, use only basename
    radical = checking.parts[-1]

    def is_under(root, potential_subdir):
        return root in potential_subdir.parents

    home = Path.home()
    # main locations where to start searching
    anchors = [
        Path.cwd(),
        Path.cwd().parent,
        Path.cwd().parent.parent,
        Path.home()
    ]
    # the subdirs where to search from the anchors
    def iter_derivatives(path):
        yield path
        yield path / "yaml"
        yield path / ".yaml"
        yield path / ".quiz"

    # possible extensions added to the radical
    extensions = ["", ".yaml", ".yml", ".yamlb" ]

    # search it
    anchors = [ path for path in anchors if is_under(home, path) ]

    paths = [ d for path in anchors for d in iter_derivatives(path)]

    failed = []
    for path in paths:
        for extension in extensions:
            candidate = path / (radical + extension)
            if candidate.exists():
                return candidate
            elif debug:
                failed.append(candidate)

    print(f"could not spot quiz {filename_or_path}")
    if debug:
        for candidate in failed:
            print(f"have tried {candidate}")


def run_yaml_quiz(filename_or_path, exoname, debug=False):
    """
    one-liner convenience helper function, and main entry point

    in a single pass:
    * locate quiz from just its name, see heuristics above
    * parse the file (or Path)
    * build a Quiz object corresponding to exoname
    * create and return widget

    use debug=True until you get the YAML file right
    """

    actual_path = locate_from_radical(filename_or_path, debug)
    if not actual_path:
        raise ValueError(f"could not spot quiz {filename_or_path}")

    try:
        loader = YamlLoader(actual_path)
        loader.rain_check()
        return loader.build_quiz(exoname, debug).widget()
    except yaml.parser.ParserError as exc:
        print(f"Could not parse {filename_or_path}\n{exc}")
    except Exception as exc:
        print(f"OOPS - something wrong with quiz {type(exc)}, {exc}")
        debug = True
        if debug:
            import traceback
            print(traceback.format_exc())
        return exc



# -*- coding: utf-8 -*-

# As a side note, I also have a simple tool for collecting corrections
# and publish them in pdf format.
# This is what the @BEG@ / @END@ thingy is for, just ignore this for now


# importing the ExerciseFunction class
from nbautoeval import ExerciseFunction, Args
from nbautoeval import CallRenderer, PPrintRenderer


#
# this is a regular Python module, so I can define
# any variable, here a global that I need in this context
# 
nucleotides = 'CAGT'



########## step 1
# You need to define the 'correct' function
# i.e. one that would be accepted as valid for the problem

# @BEG@ name=percentages
def percentages(adn):
    "calcule des percentages de CAGT dans un adn"
    total = len(adn)
    return {
        nucleotide : len([p for p in adn if p == nucleotide])*100./total
        for nucleotide in nucleotides
    }
# @END@



########## step 2
# You need to provide datasets
# This is expected to be a list of Args instances
# each one describes all the arguments to be passed
# to the function
# in this particular case we define 2 input sets, so
# the correction will have 2 meaningful rows
#
inputs_percentages = [
    Args('ACGTACGA'),
    Args('ACGTACGATCGATCGATGCTCGTTGCTCGTAGCGCT'),
]

# The Args object takes eactly the arguments to be passed to the function
# so if the function was expecting 2 arguments, then I would have added
# Args(10, 12)
# or even
# Args(10, *(12, 13, 14))



########## step 3
# finally we create the exercise object
# NOTE: this is the only name that should be imported from this module
#
exo_percentages = ExerciseFunction(
    # first argument is the 'correct' function
    # it is recommended to use the same name as in the notebook, as the
    # python function name is used in HTML rendering 
    percentages,
    # the inputs
    inputs_percentages,
    result_renderer=PPrintRenderer(width=20),
)



#
# an alternative exercise instance to demonstrate
# how to tweak the way the various columns are rendered
# an ExerciseFunction instance can provide:
# 
# * call_renderer: how to display the call (leftmost column)
# * result_renderer: how to display the results (2 middle columns)
# 
# the former should be some instance of one of the *CallRenderer classes
#   as defined in call.py
# the latter shoud be some instance of any of the *Renderer classes
#   as defined in renderer.py
# 
exo_percentages2 = ExerciseFunction(
    percentages, inputs_percentages,
    # show function name in leftmost column
    call_renderer=CallRenderer(show_function=False),
    # use pprint to format results
    result_renderer=PPrintRenderer(width=20),
)

#
# illustration on how to set fonts sizes
exo_percentages3 = ExerciseFunction(
    percentages, inputs_percentages,
    result_renderer=PPrintRenderer(width=20),
    # here
    font_size="75%",
    header_font_size="200%",
)



import random
import time

from typing import List, Union
from enum import Enum

from ipywidgets import Layout, HBox, VBox, Checkbox, Button, HTML, HTMLMath

from .content import TextContent, MarkdownMathContent, CssContent
from .storage import log2_quiz, storage_read, storage_save
from .helpers import truncate

CSS = """
:root {
    /* these two are for unanswered */
    --question-bg-odd: #d6e4f0;
    --question-bg-even: #ddebf8;
    --question-bg-right: #dafcf0;
    --question-bg-partial: #ffd6d9; /* pale pink/red */
    --question-bg-wrong: #c94277;      /* darker red */
    /* individual options after revealed */
    --explanation-right: #f8f8f8;
    --explanation-wrong: #f0f0f0;
    /* misc */
    --question-header-bg: white;
    --submit-bg: #c2f0fc;
    /* borders */
    --border-question: 2px solid #084177;
    --border-code: 1px solid #a0a0a0;
    --border-explanation: 1px solid #b0b0b0;
    --border-submit: 2px solid #084177;
    --border-separator-options: 0.8px solid navy;
}

.widget-vbox.nbae-question, .widget-hbox.nbae-question {
    padding: 10px;
    border-radius: 10px;
    margin-bottom: 5px;
}
.nbae-question .question {
    border: var(--border-question);
    border-radius: 6px;
    width: max-content;
    max-width: 100%;
    padding: 4px 8px;
    background-color: var(--question-header-bg);
}
/* question in a QuizQuestion with a vertical layout */
.nbae-question.widget-vbox>.question {
    width: 100%;
}

.nbae-question .index {
    font-weight: bold;
}

.nbae-question .question:not(.exactly-one) .question-header::after {
    content: "♧";
    align-self: center;
    font-weight: bold;
    font-size: 125%;
}
.nbae-question .question-header {
    padding-top: 8px;
    padding-bottom: 8px;
    border-bottom: 1px solid #084177;
}
.nbae-question .score::before {
    content: "|";
    align-self: center;
    margin-right: 10px;
}
.nbae-question .score {
    margin-left: 10px;
    margin-right: 10px;
}

.nbae-question .code pre {
    line-height: 1.3;
    padding: 5px;
}
.nbae-question hr {
    height: 1px;
    background-color: #202020;
}
.nbae-question .code {
    border: var(--border-code);
}

.nbae-question .widget-checkbox {
    width: auto;
    padding-left: 10px;
}

.nbae-question .question-sequel {
    width: auto;
    margin-left: 10px;
    padding: 8px 10px;
    /*border: var(--border-question);*/
    background-color: var(--question-header-bg);
    border-radius: 6px;
}

.nbae-question.right, .nbae-quiz .summary.ok {
    background-color: var(--question-bg-right);
}
/* keep this light pink and not dark red */
.nbae-question.wrong, .nbae-quiz .summary.ko {
    background-color: var(--question-bg-partial);
}
.nbae-question.partial {
    background-color: var(--question-bg-partial);
}
.nbae-question.unanswered {
    background-color: var(--question-bg-odd);
}
.nbae-question.unanswered:nth-child(2n) {
    background-color: var(--question-bg-even);
}

.nbae-quiz .submit {
    margin: 10px;
    border-radius: 10px;
    background-color: var(--submit-bg);
    border: var(--border-submit);
    width: max-content;
}

.nbae-quiz .result-area {
    align-items: center;
}

.nbae-quiz .summary {
    padding: 5px 20px;
    width: fit-content;
}

.nbae-quiz .wrong-answer {
    background-color: var(--question-bg-wrong);
    border-radius: 6px;
}

.nbae-quiz .main-score {
    font-weight: bold;
    font-size: larger;
    padding-left: 6px;
    padding-right: 6px;
}

.nbae-question div.score span {
    font-weight: bold;
    font-size: larger;
    padding: 4px;
}

.nbae-question.unanswered span.unanswered {
    background-color: var(--question-bg-odd);
}
.nbae-question.unanswered:nth-child(2n) span.unanswered {
    background-color: var(--question-bg-even);
}
.nbae-question.right span.right {
    background-color: var(--question-bg-right);
}
.nbae-question.wrong span.wrong {
    background-color: var(--question-bg-wrong);
}
.nbae-question.partial span.partial {
    background-color: var(--question-bg-partial);
}

.nbae-question div.options>div.wrong-answer {
    margin-top: 8px;
}
.nbae-question div.options.widget-vbox>div:not(:last-child) {
    border-bottom: var(--border-separator-options);
    padding-bottom: 5px;
}
.nbae-question div.options.widget-hbox>div:not(:last-child) {
    border-right: var(--border-separator-options);
    padding-right: 5px;
}

.nbae-question .explanation::before {
    content: "☛";
    margin-right: 10px;
    font-size: 150%;
}
.nbae-question .explanation {
    border-radius: 6px;
    padding: 10px;
    background-color: var(--explanation-right);
    display: none;
    border: var(--border-explanation);
}

.nbae-question .wrong-answer .explanation {
    background-color: var(--explanation-wrong);
}

.nbae-question:not(.revealed) .explanation {
    display: none;
}
.nbae-question.revealed .explanation {
    display: inherit;
}
.nbae-question /*div.widget-html-content*/ pre {
    line-height: 1.2;
}
.nbae-question div.widget-html-content {
    line-height: 1.6;
}
.nbae-question div.option-box div {
    align-self: center;
}
/* from https://stackoverflow.com/questions/8206315/css3-tooltip-with-hoverafter-positioning-and-size */
*[tooltip] {
  position: relative;
  border-bottom: dotted 1px #000;
}
*[tooltip]:hover:before {
/**[tooltip]:before {*/
  content: attr(tooltip);
  background-color: #000;
  color: #fff;
  /*top: 1em;*/
  /*bottom: 100%;*/
  left: 105%;
  position: absolute;
  white-space: nowrap;
  padding: 5px;
  border-radius: 4px;
  z-index: 1000;
  font-weight: initial;
  font-size: initial;
}
"""

def now():
    return dict(date=time.strftime("%Y-%m-%d"),
                time=time.strftime("%H:%M:%S"))

# unless specified otherwise in yaml
DEFAULT_CONTENT_CLASS = MarkdownMathContent


# a flexible content can be defined either with a plain str
# or with a Content object
FlexibleContent = Union[str, TextContent]

class Flexible:
    # it's important for the YAML loader that this parameter be called text
    def __init__(self, text: FlexibleContent):
        if isinstance(text, TextContent):
            self.content = text
        elif isinstance(text, str):
            self.content = DEFAULT_CONTENT_CLASS(text)
        else:
            raise ValueError(f"unexpected type {type(text).__name__}"
                             f" for flexible {text}")
    def widget(self):
        return self.content.widget()


class Explanation(Flexible):
    def widget(self):
        w = super().widget()
        return w.add_class('explanation') if w else None


class GenericBooleanOption:
    """
    an Option instance represents one of the possible answers
    by default it is not a correct answer
    """
    def __init__(self, *, correct=False):
        self.correct = correct
        self.selected = None
    def render(self):
        print(f"Option classes must implement render()")


class Option(GenericBooleanOption):
    """
    the most basic kind of Option allows to enter a simple text
    see more specialized classes for other kinds of inputs,
    like CodeOption, MathOption and MarkdownOption

    correct=True should be set on options that are valid and
    that students should check as such

    """
    def __init__(self, text, explanation=None, **kwds):
        super().__init__(**kwds)
        self.text = text
        self.explanation = explanation
    def render(self):
        return TextContent(self.text)
    def explanation_widget(self):
        return None if not self.explanation else self.explanation.widget()


class CodeOption(Option):
    def render(self):
        return super().render().set_is_code(True).add_class('code')

class MathOption(Option):
    def render(self):
        return super().render().set_needs_math(True)

class MarkdownOption(Option):
    def render(self):
        return super().render().set_has_markdown(True)

class MarkdownMathOption(Option):
    def render(self):
        return super().render().set_has_markdown(True).set_needs_math(True)

DEFAULT_OPTION_CLASS = MarkdownMathOption


# this class captures the order in which options are provided
# in the QuizQuestion object
class _TeacherOptions:
    def __init__(self, options: List[GenericBooleanOption]):
        self.options = options
    def append(self, option_none):
        self.options.append(option_none)
    def __iter__(self):
        return iter(self.options)


# this class captures the order in which options or questions
# actually displayed, which is randomized from the input list
# (as defined in the YAML code) when shuffle is True
class _DisplayedItems:
    def __init__(self, input_list, shuffle):
        self.displayed = input_list[:]
        if shuffle:
            random.shuffle(self.displayed)
    def append(self, option_none):
        self.displayed.append(option_none)
    def __iter__(self):
        return iter(self.displayed)

class _DisplayedOptions(_DisplayedItems):
    def correct_indices(self):
        return [i for (i, opt) in enumerate(self.displayed) if opt.correct]

class _DisplayedQuestions(_DisplayedItems):
    pass



class Answer(Enum):
    UNANSWERED = -1   # MUST BE <0
    WRONG = 0
    RIGHT = 1000


# cosmetic, show integers as :d and reals as :.2f
def display(score):
    if abs(score-round(score)) < 0.01:
        return f"{round(score):d}"
    else:
        return f"{score:.2f}"

def point_or_points(score):
    return f"{display(score)} {'pt' if score<=1 else 'pts'}"

# to allow for QuizQuestion(... score=1)
# or           QuizQuestion(... score=(6, -1))
# or           QuizQuestion(... score=(6, -3, 1))
class Score:
    def __init__(self, single_arg=None):
        self.if_right = 1
        self.if_wrong = -1
        self.if_unanswered = 0
        # accept lists for the yaml loader
        ok_types = (tuple, list)
        if single_arg is None:
            pass
        elif isinstance(single_arg, int):
            self.if_right = single_arg
        elif isinstance(single_arg, ok_types) and len(single_arg) == 2:
            self.if_right, self.if_wrong = single_arg
        elif isinstance(single_arg, ok_types) and len(single_arg) == 3:
            self.if_right, self.if_wrong, self.if_unanswered = single_arg
        else:
            raise ValueError(f"Score constructor expects a single int\n"
                             f"or a {ok_types} of 2 or 3 ints")
        # this is the default; use use_progressive_grading()
        # to set otherwise, not in the constructor signature
        # because of the way the loader works
        self.all_or_nothing = True

    def use_progressive_grading(self):
        self.all_or_nothing = False


    def all_or_nothing_score(self, answer):
        return (self.if_right if answer == Answer.RIGHT
                else self.if_wrong if answer == Answer.WRONG
                else self.if_unanswered)

    def progressive_score(self, nb_options, nb_correct_answers):
        if nb_correct_answers == Answer.UNANSWERED:
            return self.if_unanswered
        if nb_correct_answers == Answer.RIGHT:
            return self.if_right
        return (self.if_wrong
               + (nb_correct_answers/nb_options)*(self.if_right-self.if_wrong))

    def html(self, partial_score="‒‒", partial_message=None):
        def blob(klass, text, msg):
            return f"<span class='{klass}' tooltip='{msg}'>{text}</span>"
        result =  ""
        right_points = point_or_points(self.if_right)
        result += blob('right', right_points, f" {right_points} for a correct answer")
        wrong_points = point_or_points(self.if_wrong)
        if not self.all_or_nothing:
            result += " / "
            message = (partial_message or
                       f"between {wrong_points} and {right_points} "
                       f"for a partially good answer")
            result += blob('partial', partial_score, message)
        result += " / "
        result += blob('wrong', wrong_points, f"{wrong_points} for a wrong answer")
        result += " / "
        unanswered_points = point_or_points(self.if_unanswered)
        result += blob('unanswered', unanswered_points,
                       f"{unanswered_points} if not answered at all")
        return result

    def __str__(self):
        return f"{self.if_right}/{self.if_wrong}/{self.if_unanswered}"


class QuizQuestion:
    """
    question can be a str, or a TextContent object for more complex inputs;
    it may include html tags and/or math content between '$$'

    options is a list of Option objects; if exactly_one_option is set,
    then obviously exactly one of these options must have correct=True
    (this needs to be explicit, an options list with one correct option
    is not deemed enough a condition)
    when exactly_one_option is set, the checkboxes behave like radio buttons

    if option_none is provided, it should be an Option object, that will
    be guaranteed to appear last even when options are shuffled; it is designed
    so that a teacher can create a 'none of the above' option

    shuffle is a boolean indicating if the options must be shuffled
    around for each student

    when horizontal_layout is set, the answers appear on the right of the question,
    otherwise they appear below

    when horizontal_options is set, the answers are displayed in a horizontal box
    instead of a vertical one
    """

    # constructor is kept as slight as possible for the yaml loader
    # in that context options is not yet a list of Options objects
    # but just plain Python dicts as outcome from yaml
    def __init__(self, *,
                 question: FlexibleContent,
                 options: List,
                 # if defined, show up on top of the alternatives
                 question_sequel: FlexibleContent=None,
                 explanation=None,
                 # do we want to shuffle the options
                 shuffle=True,
                 # set this to True to mak it plain
                 # that there is exactly one option to select
                 exactly_one_option=False,
                 all_or_nothing=None,
                 option_none=None,
                 # how to display
                 # for now, this is simple
                 score=1,
                 horizontal_layout=False,
                 horizontal_options=False):
        self.question = question
        self.options = options
        self.question_sequel = question_sequel
        self.explanation = explanation
        self.shuffle = shuffle
        self.exactly_one_option = exactly_one_option
        # default if unspecified depends on type
        if all_or_nothing is not None:
            self.all_or_nothing = all_or_nothing
        else:
            self.all_or_nothing = self.exactly_one_option
        # default for scores historically is all_or_nothing
        self._score_object = Score(score)
        if not self.all_or_nothing:
            self._score_object.use_progressive_grading()
        self.horizontal_layout = horizontal_layout
        self.horizontal_options = horizontal_options
        self.option_none = option_none
        #
        self.feedback_area = None
        self._widget_instance = None
        # the rank in the Quiz objectf.explanation
        self.index = None
        #
        self._post_inited = False

    @property
    def nb_options(self):
        return len(self.options)

    # post-processing once options has been elaborated
    def post_init(self):
        if self._post_inited:
            return
        # shuffle if requested
        self._displayed_options = _DisplayedOptions(self.options, self.shuffle)
        # add 'none of the above' option as last option if requested
        if self.option_none is not None:
            self.options.append(self.option_none)
            self._displayed_options.append(self.option_none)
        self.sanity_check()
        self._post_inited = True


    def sanity_check(self):
        def report(*messages):
            print(f"**WARNING** : question `{truncate(str(self.question), 70)}`\n\t", *messages)
        nb_correct_options = len(self._displayed_options.correct_indices())
        if self.exactly_one_option:
            if  nb_correct_options != 1:
                report(f"has {nb_correct_options} correct answers, expected 1 b/c exactly_one_option")
        else:
            if  nb_correct_options == 0:
                report(f"has no correct answers, this is not supported,\n"
                       f"\tplease add an option like 'none of the other answers'")


    def set_index(self, index):
        self.index = index


    def detailed_answer(self):
        """
        * if self.all_on_nothing
          returns Answer.UNANSWERED or Answer.RIGHT or Answer.WRONG

        * otherwise, for progressive grading, returns
          Answer.UNANSWERED or Answer.WRONG if all options are wrong
          or the number of correct options if not quite right
          or Answer.RIGHT if all options are OK

        """

        selected = [i for (i, checkbox) in enumerate(self.checkboxes)
                    if checkbox.value]
        if not selected:
            # importantly UNANSWERED is -1
            return Answer.UNANSWERED
        if self.all_or_nothing:
            return (Answer.RIGHT
                    if set(selected) == set(self._displayed_options.correct_indices())
                    else Answer.WRONG)
        else:
            world = set(range(self.nb_options))
            should_be_checked = set(self._displayed_options.correct_indices())
            should_be_unchecked = world - should_be_checked
            are_checked = set(selected)
            are_unchecked = world - are_checked
            right_options = (len(should_be_checked & are_checked)
                            +len(should_be_unchecked & are_unchecked))
            if right_options == self.nb_options:
                return Answer.RIGHT
            else:
                return right_options


    def score(self):
        if self.all_or_nothing:
            return self._score_object.all_or_nothing_score(
                self.detailed_answer())
        else:
            return self._score_object.progressive_score(
                self.nb_options, self.detailed_answer()
            )


    def max_score(self):
        return self._score_object.if_right


    def widget(self):

        self.post_init()

        if self._widget_instance:
            return self._widget_instance

        # header area
        self._score_widget = HTML(f'{self._score_object.html()}').add_class('score')
        header_widget = HBox([
            HTML(f'Question # {self.index}').add_class('index'),
            self._score_widget,
        ]).add_class('question-header')
        question_widget = Flexible(self.question).widget()
        header_widgets = [header_widget, question_widget]
        if self.explanation:
            header_widgets.append(self.explanation.widget())
        question = VBox(header_widgets).add_class('question')
        if self.exactly_one_option:
            question.add_class('exactly-one')

        # the options per se
        # it's important that we have as many checkboxes as option_boxes
        self.checkboxes = [Checkbox(value=option.selected, disabled=False,
                                    description='', indent=False)
                           for option in self._displayed_options]
        # arm callback so that choices get saved at all times
        # xxx not working at this point as we have no means
        # to retrieve the parent Quiz object to trigger a save_preserved
        #def preserve(b):
        #    self.xxx_quiz_xxx.save_preserved()
        for checkbox in self.checkboxes:
            # checkbox.observe(preserve)
            # radio-box behaviour
            if self.exactly_one_option:
                checkbox.observe(lambda event: self.radio_button_callback(event))
        labels = [option.render().widget() for option in self._displayed_options]

        # explanations contains, for each option, either a widget, or None
        explanations = [option.explanation_widget()
                             for option in self._displayed_options]

        options_box = HBox if self.horizontal_options else VBox
        def make_option(checkbox, label, explanation):
            main = HBox([checkbox, label]).add_class('option-box')
            return (main if not explanation
                    else VBox([main, explanation]))
        self.option_boxes = [
            make_option(checkbox, label, explanation)
            for (checkbox, label, explanation)
            in zip(self.checkboxes, labels, explanations)]
        if not self.question_sequel:
            actual_sons = self.option_boxes
        else:
            actual_sons = [Flexible(self.question_sequel).widget()
                           .add_class('question-sequel')]
            actual_sons += self.option_boxes
        options = options_box(actual_sons)
        options.add_class('options')

        css_widget = CssContent(CSS).widget()

        # putting it all together
        layout_box = HBox if self.horizontal_layout else VBox
        self._widget_instance = layout_box(
            [question, options, css_widget])
        self._widget_instance.add_class('nbae-question')
        self.feedback_area = self._widget_instance
        self.feedback(Answer.UNANSWERED)
        return self._widget_instance


    # this will be bound to the checkbox widgets through observe()
    # which passes along an event object
    def radio_button_callback(self, event):
        # only interested in that sort of events
        if event['name'] != 'value':
            return
        # locate the checkbox that is being changed
        checkbox = event['owner']
        # we only react on changes that set value=True
        if checkbox.value == False:
            return
        if checkbox not in self.checkboxes:
            print("WHOOPS - s/t wrong !")
            return
        for other in self.checkboxes:
            if other is not checkbox:
                other.value = False


    def feedback(self, answer: Answer):
        """
        assuming the widget was created already, of course
        """
        all_classes = ('right', 'partial', 'wrong', 'unanswered')
        if self.feedback_area is None:
            return
        if answer == Answer.UNANSWERED:
            on = 'unanswered'
        elif answer == Answer.RIGHT:
            on = 'right'
        elif answer == Answer.WRONG:
            on = 'wrong'
        else:
            on = 'partial'
        self.feedback_area.add_class(on)
        for off in all_classes:
            if off != on:
                self.feedback_area.remove_class(off)
        # the score area
        if self.all_or_nothing:
            return
        if on != 'partial':
            self._score_widget.value = self._score_object.html()
        else:
            right_options = self.detailed_answer()
            partial_score = self._score_object.progressive_score(
                self.nb_options, right_options)
            message = (f"partial score based on "
                       f"{right_options}/{self.nb_options} good answers")
            self._score_widget.value = self._score_object.html(
                point_or_points(partial_score), message)


    def individual_feedback(self):
        for option, option_box, checkbox in zip(
            self._displayed_options, self.option_boxes, self.checkboxes):
            checkbox.disabled = True
            # good answer ?
            if option.correct == checkbox.value:
                option_box.remove_class('wrong-answer')
            else:
                option_box.add_class('wrong-answer')
        self._widget_instance.add_class('revealed')


    def preserve(self) -> List[bool]:
        for option, checkbox in zip(self._displayed_options, self.checkboxes):
            option.selected = checkbox.value
        return [option.selected for option in self.options]


    def restore(self, bools: List[bool]):
        for option, boolean in zip(self.options, bools):
            option.selected = boolean
        if self._widget_instance:
            for checkbox, option in zip(self.checkboxes, self._displayed_options):
                checkbox.value = option.selected


class Quiz:
    """
    a quiz is made of several questions
    one can only submit a full Quiz, not just one question at a time
    """

    # same approach to hyper-light constructor
    # the questions attribute might temporarily be a list of
    # plain Python objects, not QuizQuestion instances yet
    def __init__(self,
                 *,
                 exoname,
                 questions: List[QuizQuestion],
                 shuffle=True,
                 max_attempts=2,
                 max_grade=None):
        self.exoname = exoname
        self.questions = questions
        self.shuffle = shuffle
        self.max_attempts = max_attempts
        self.max_grade = max_grade

        # private - for updating the UI
        self.submit_button = None
        self.submit_summary = None
        #
        self._post_inited = False


    def post_init(self):
        if self._post_inited:
            return
        # this is tricky and not quite right
        # we need to have the questions post-init'ed
        # this early because of the restoration below
        # otherwise questions won't have their option_none appended
        # and restore would fail to restore the option_none option
        for question in self.questions:
            question.post_init()
        self.displayed_questions = _DisplayedQuestions(self.questions, self.shuffle)
        # needs to be saved somewhere
        self.current_attempts = storage_read(self.exoname, 'current_attempts', 0)
        preserved = storage_read(self.exoname, "preserved", [])
        if preserved:
            self.restore(preserved)
        # set question rank
        for index, question in enumerate(self.displayed_questions, 1):
            question.set_index(index)
        self._post_inited = True


    def widget(self):
        self.post_init()
        sons = [question.widget() for question in self.displayed_questions]

        self.submit_button = Button(description='submit').add_class('submit')
        self.submit_summary = HTML('no result yet').add_class('summary')
        self.submit_button.on_click(lambda button: self.submit(button))
        sons.append(HBox([self.submit_button, self.submit_summary])
                    .add_class('result-area'))
        toplevel = VBox(sons).add_class('nbae-quiz')
        self.update(within_submit=False)
        return toplevel


    def submit(self, _button):
        self.current_attempts += 1
        # because of possible load/network latency,
        #  we need to disable this until the event is processed
        # update() will re-enable it later on if needed
        self.submit_button.disabled = True
        self.update(within_submit=True)
        storage_save(self.exoname, 'current_attempts', self.current_attempts)
        self.save_submitted()
        # xxx no longer needed if we can get all changes to be saved
        self.save_preserved()
        (current_score, max_score,
         normalized_score, normalized_max_score) = self.total_score()
        log_kwds = {}
        if self.max_grade:
            log_kwds = dict(normalized_score=normalized_score,
                            normalized_max_score=normalized_max_score)
        log2_quiz(self.exoname,
                  attempt=self.current_attempts, max_attempts=self.max_attempts,
                  score=current_score, max_score=max_score,
                  **log_kwds)


    def preserve(self) -> List[List[bool]]:
        return [question.preserve() for question in self.questions]

    def restore(self, list_of_list_of_bools):
        for question, list_of_bools in zip(self.questions, list_of_list_of_bools):
            question.restore(list_of_bools)


    def save_preserved(self):
        storage_save(self.exoname, "preserved", self.preserve())
    def save_submitted(self):
        history = storage_read(self.exoname, "submitted", [])
        this_attempt = dict(when=now(), answers=self.preserve(), html=self.summary_html)
        history.append(this_attempt)
        storage_save(self.exoname, "submitted", history)


    # final no longer used
    def score_html(self, right_answers):
        s, m, ns, nm = self.total_score()
        score = ""
        score += (f"score for attempt #{self.current_attempts}:&nbsp;&nbsp;&nbsp; ")
        if self.max_grade is None:
            score += f"<span class='main-score'>{display(s)}/{m}</span>"
        else:
            score += f"{display(s)}/{m}"
            score += f" = <span class='main-score'>{display(ns)}/{nm}</span>"
        score += f" ({len(right_answers)}/{len(self.answers)} questions OK)"
        return score


    def update(self, *, within_submit):
        # there may be latency if the host is loaded
        self.answers = [question.detailed_answer()
                        for question in self.displayed_questions]
        right_answers = [answer for answer in self.answers if answer == Answer.RIGHT]
        all_right = (len(right_answers) == len(self.answers))

        history = storage_read(self.exoname, "submitted", [])
        present = now()
        today, right_now = present['date'], present['time']
        def past_entry(event):
            when = event['when']
            if when['date'] == today:
                timestamp = f"@{when['time']}"
            else:
                timestamp = f"on{when['date']} @{when['time']}"
            return f"{timestamp} {event['html']}"
        past_summaries = [
            past_entry(event) for event in history
        ]
        def update_summary(current_summary):
            aggregate = ""
            aggregate += "<div>"
            for summary in past_summaries:
                aggregate += f"{summary}<br>"
            if within_submit:
                aggregate += f"@{right_now} {current_summary}"
            aggregate += "</div>"
            self.submit_summary.value = aggregate
        # quiz is over - either way (all good or attempts ran out)
        if all_right or self.current_attempts >= self.max_attempts:
            # materialize all questions
            for question in self.displayed_questions:
                question.feedback(question.detailed_answer())
                question.individual_feedback()
            self.submit_button.description = "quiz over"
            self.summary_html = f"{self.score_html(right_answers)}"
            update_summary(self.summary_html)
            if all_right:
                self.submit_summary.add_class('ok')
            else:
                self.submit_summary.add_class('ko')
            # not really needed (it's been disable already in submit)
            # but for consistency
            self.submit_button.disabled = True
        else:
            # so here we know that self.current_attempts < self.max_attempts:
            left = self.max_attempts - self.current_attempts
            self.submit_button.description = (
                f"submit ({left}/{self.max_attempts} attempts left)")
            if self.current_attempts >= 1:
                self.summary_html = f"{self.score_html(right_answers)}"
                update_summary(self.summary_html)
            self.submit_button.disabled = False

    def total_score(self):
        """
        returns a tuple current_score, max_score
        """
        current_score = sum(q.score() for q in self.questions)
        max_score = sum(q.max_score() for q in self.questions)
        if self.max_grade is not None:
            normalized_max_score = self.max_grade
            normalized_score = current_score/max_score * normalized_max_score
        else:
            normalized_max_score, normalized_score = max_score, current_score
        return current_score, max_score, normalized_score, normalized_max_score
